"""parses [PREDICT] section of config"""
from configparser import NoOptionError

import attr
from attr import converters, validators
from attr.validators import instance_of, optional

from .converters import comma_separated_list, expanded_user_path
from .validators import is_a_directory, is_a_file, is_valid_model_name


@attr.s
class PredictConfig:
    """class that represents [PREDICT] section of config.ini file

    Attributes
    ----------
    csv_path : str
        path to where dataset was saved as a csv.
    checkpoint_path : str
        path to directory with checkpoint files saved by Tensorflow, to reload model
    models : list
        of model names. e.g., 'models = TweetyNet, GRUNet, ConvNet'
    spect_scaler_path : str
        path to a saved SpectScaler object used to normalize spectrograms.
        If spectrograms were normalized and this is not provided, will give
        incorrect results.
    """
    # required
    checkpoint_path = attr.ib(converter=expanded_user_path,
                              validator=is_a_directory)
    models = attr.ib(converter=comma_separated_list,
                     validator=[instance_of(list), is_valid_model_name])

    # optional
    # csv_path is actually 'required' but we can't enforce that here because cli.prep looks at
    # what sections are defined to figure out where to add csv_path after it creates the csv
    csv_path = attr.ib(converter=converters.optional(expanded_user_path),
                       validator=validators.optional(is_a_file),
                       default=None
                       )

    spect_scaler_path = attr.ib(validator=optional([instance_of(str), is_a_file]),
                                default=None)


REQUIRED_PREDICT_OPTIONS = [
    'checkpoint_path',
    'models',
]


def parse_predict_config(config_obj, config_path):
    """parse [PREDICT] section of config.ini file

    Parameters
    ----------
    config_obj : ConfigParser
        containing config.ini file already loaded by parse function
    config_path : str
        path to config.ini file (used for error messages)

    Returns
    -------
    predict_config : vak.config.predict.PredictConfig
        instance of PredictConfig class that represents [PREDICT] section
        of config.ini file
    """
    predict_section = dict(
        config_obj['PREDICT'].items()
    )

    for required_option in REQUIRED_PREDICT_OPTIONS:
        if required_option not in predict_section:
            raise NoOptionError(
                f"the '{required_option}' option is required but was not found in the "
                f"PREDICT section of the config.ini file: {config_path}"
            )
    return PredictConfig(**predict_section)



class OnlyValDurError(Exception):
    pass


class InvalidDurationError(Exception):
    pass


class SplitsDurationGreaterThanDatasetDurationError(Exception):
    pass


def _validate_durs(train_dur, val_dur, test_dur, vds_dur):
    """helper function to validate durations specified for splits

    If train_dur, val_dur, and test_dur are all None, a ValueError is raised.

    If any of train_dur, val_dur, or test_dur have a negative value that is not -1, an
    InvalidDurationError is raised. -1 is interpreted differently as explained below.

    If all three have non-negative values, this function simply checks that their sum is not
    greater than vds_dur. If this is True, it returns them unchanged. If the total sum *is*
    greater than vds_dur, an error is raised (SplitsDurationGreaterThanDatasetDurationError).

    If only val_dur is specified, this raises a ValDurError; not clear what durations of training
    and test set should be.

    If only train_dur is specified, then test_dur is set to -1; similarly if only test_dur is
    specified, then train_dur is set to -1. Other functions interpret this as "first get the
    split for the set with a value specified, then use the remainder of the dataset in the split
    whose duration is set to -1".

    Parameters
    ----------
    train_dur : int, float
        Target duration for training set, in seconds.
    val_dur : int, float
        Target duration for validation set, in seconds.
    test_dur : int, float
        Target duration for test set, in seconds.
    vds_dur : int, float
        Total duration of Dataset.

    Returns
    -------
    train_dur, val_dur, test_dur : int, float
    """
    if all([dur is None for dur in (train_dur, val_dur, test_dur)]):
        raise ValueError("train_dur, val_dur, and test_dur were all None; must specify at least train_dur or test_dur")

    else:
        if not all([dur >= 0 or dur == -1 for dur in (train_dur, val_dur, test_dur) if dur is not None]):
            raise InvalidDurationError("all durations for split must be real non-negative number or "
                                       "set to -1 (meaning 'use the remaining dataset)")

        if val_dur and train_dur is None and test_dur is None:
            raise OnlyValDurError(
                'cannot specify only val_dur, unclear how to split dataset into training and test sets'
            )

        if train_dur:
            if (test_dur is None and val_dur is None) or (val_dur and test_dur is None):
                test_dur = -1  # keep val_dur None

        elif test_dur:  # and train_dur was None
            if (train_dur is None and val_dur is None) or (val_dur and train_dur is None):
                train_dur = -1  # keep val_dur None

        if -1 not in (train_dur, val_dur, test_dur):
            total_splits_dur = sum([dur for dur in (train_dur, val_dur, test_dur) if dur is not None])
            if total_splits_dur > vds_dur:
                raise SplitsDurationGreaterThanDatasetDurationError(
                    f'total of durations specified for dataset split, {total_splits_dur} s, '
                    f'is greater than total duration of Dataset, {vds_dur}.'
                )

    return train_dur, val_dur, test_dur



import numpy as np


def standardize_spect(spect, mean_freqs, std_freqs, non_zero_std):
    """standardize spectrogram by subtracting off mean and dividing by standard deviation.

    Parameters
    ----------
    spect : numpy.ndarray
        with shape (frequencies, time bins)
    mean_freqs : numpy.ndarray
        vector of mean values for each frequency bin across the fit set of spectrograms
    std_freqs : numpy.ndarray
        vector of standard deviations for each frequency bin across the fit set of spectrograms
    non_zero_std : numpy.ndarray
        boolean, indicates where std_freqs has non-zero values. Used to avoid divide-by-zero errors.

    Returns
    -------
    transformed : numpy.ndarray
        with same shape as spect but with (approximately) zero mean and unit standard deviation
        (mean and standard devation will still vary by batch).
    """
    tfm = spect - mean_freqs[:, np.newaxis]  # need axis for broadcasting
    # keep any stds that are zero from causing NaNs
    tfm[non_zero_std, :] = tfm[non_zero_std, :] / std_freqs[non_zero_std, np.newaxis]
    return tfm



from functools import partial
import json
import logging
from pathlib import Path
import shutil
import sys
from datetime import datetime

import joblib
import pandas as pd
import torch.utils.data
from torchvision import transforms

from .. import config
from .. import models
from .. import util
from ..datasets.spectrogram_window_dataset import SpectrogramWindowDataset
from ..io import dataframe
from ..transforms import StandardizeSpect


def train(toml_path):
    """train models using training set specified in config.toml file.
    Function called by command-line interface.

    Parameters
    ----------
    toml_path : str, Path
        path to a configuration file in TOML format.

    Returns
    -------
    None

    Trains models, saves results in new directory within root_results_dir specified
    in config.ini file, and adds path to that new directory to config.ini file.
    """
    toml_path = Path(toml_path)
    cfg = config.parse.from_toml(toml_path)

    if cfg.train is None:
        raise ValueError(
            f'train called with a config.toml file that does not have a TRAIN section: {toml_path}'
        )

    dataset_df = pd.read_csv(cfg.train.csv_path)
    # ---------------- pre-conditions ----------------------------------------------------------------------------------
    if cfg.train.val_error_step and not dataset_df['split'].str.contains('val').any():
        raise ValueError(
            f"val_error_step set to {cfg.train.val_error_step} but dataset does not contain a validation set; "
            f"please run `vak prep` with a config.ini file that specifies a duration for the validation set."
        )

    # ---- set up directory to save output -----------------------------------------------------------------------------
    timenow = datetime.now().strftime('%y%m%d_%H%M%S')
    results_dirname = f'results_{timenow}'
    if cfg.train.root_results_dir:
        results_path = Path(cfg.train.root_results_dir)
    else:
        results_path = Path('.')
    results_path = results_path.joinpath(results_dirname)
    results_path.mkdir(parents=True)
    # copy config file into results dir now that we've made the dir
    shutil.copy(toml_path, results_path)

    # ---- set up logging ----------------------------------------------------------------------------------------------
    logfile_name = results_path.joinpath('logfile_from_train_' + timenow + '.log')
    logger = logging.getLogger(__name__)
    logger.setLevel('INFO')
    logger.addHandler(logging.FileHandler(logfile_name))
    logger.addHandler(logging.StreamHandler(sys.stdout))
    logger.info('Logging results to {}'.format(results_path))

    timebin_dur = dataframe.validate_and_get_timebin_dur(dataset_df)
    logger.info(
        f'Size of each timebin in spectrogram, in seconds: {timebin_dur}'
    )

    # ---------------- load training data  -----------------------------------------------------------------------------
    if cfg.train.normalize_spectrograms:
        logger.info('will normalize spectrograms')
        standardize = StandardizeSpect.fit_df(dataset_df, spect_key=cfg.spect_params.spect_key)
        joblib.dump(standardize,
                    results_path.joinpath('StandardizeSpect'))
    else:
        standardize = None

    # below, if we're going to train network to predict unlabeled segments, then
    # we need to include a class for those unlabeled segments in labelmap,
    # the mapping from labelset provided by user to a set of consecutive
    # integers that the network learns to predict
    has_unlabeled = util.dataset.has_unlabeled(cfg.train.csv_path, cfg.prep.labelset, cfg.spect_params.timebins_key)
    if has_unlabeled:
        map_unlabeled = True
    else:
        map_unlabeled = False
    labelmap = util.labels.to_map(cfg.prep.labelset, map_unlabeled=map_unlabeled)
    logger.debug(
        f'number of classes in labelmap: {len(labelmap)}'
    )
    # save labelmap in case we need it later
    with open(results_path.joinpath('labelmap.json'), 'w') as f:
        json.dump(labelmap, f)

    logger.info(f'using training dataset from {cfg.train.csv_path}')

    # make an "add channel" transform to use with Lambda
    # this way a spectrogram 'image' has a "channel" dimension (of size 1)
    # that convolutional layers can work on
    def to_floattensor(ndarray):
        return torch.from_numpy(ndarray).float()

    add_channel = partial(torch.unsqueeze, dim=0)
    transform = transforms.Compose(
        [transforms.Lambda(standardize),
         transforms.Lambda(to_floattensor),
         transforms.Lambda(add_channel)]
    )

    def to_longtensor(ndarray):
        return torch.from_numpy(ndarray).long()
    target_transform = transforms.Compose(
        [transforms.Lambda(to_longtensor)]
    )

    train_dataset = SpectrogramWindowDataset.from_csv(csv_path=cfg.train.csv_path,
                                                      split='train',
                                                      labelmap=labelmap,
                                                      window_size=cfg.dataloader.window_size,
                                                      spect_key=cfg.spect_params.spect_key,
                                                      timebins_key=cfg.spect_params.timebins_key,
                                                      transform=transform,
                                                      target_transform=target_transform
                                                      )
    train_data = torch.utils.data.DataLoader(dataset=train_dataset,
                                             shuffle=cfg.train.shuffle,
                                             batch_size=cfg.train.batch_size,
                                             num_workers=cfg.train.num_workers)
    train_dur = dataframe.split_dur(dataset_df, 'train')
    logger.info(
        f'Total duration of training set (in s): {train_dur}'
    )

    # ---------------- load validation set (if there is one) -----------------------------------------------------------
    if cfg.train.val_error_step:
        val_dataset = SpectrogramWindowDataset.from_csv(csv_path=cfg.train.csv_path,
                                                        split='val',
                                                        labelmap=labelmap,
                                                        window_size=cfg.dataloader.window_size,
                                                        spect_key=cfg.spect_params.spect_key,
                                                        timebins_key=cfg.spect_params.timebins_key,
                                                        transform=transform,
                                                        target_transform=target_transform
                                                        )
        val_data = torch.utils.data.DataLoader(dataset=val_dataset,
                                               shuffle=False,
                                               batch_size=cfg.train.batch_size,
                                               num_workers=cfg.train.num_workers)
        val_dur = dataframe.split_dur(dataset_df, 'val')
        logger.info(
            f'Total duration of validation set (in s): {val_dur}'
        )

        logger.info(
            f'will measure error on validation set every {cfg.train.val_error_step} steps of training'
        )
    else:
        val_data = None

    model_config_map = config.models.map_from_path(toml_path, cfg.train.models)
    models_map = models.from_model_config_map(
        model_config_map,
        num_classes=len(labelmap),
        input_shape=train_dataset.shape
    )
    for model_name, model in models_map.items():
        results_model_root = results_path.joinpath(model_name)
        results_model_root.mkdir()
        ckpt_root = results_model_root.joinpath('checkpoints')
        ckpt_root.mkdir()
        logger.info(
            f'training {model_name}'
        )
        model.fit(train_data=train_data,
                  num_epochs=cfg.train.num_epochs,
                  ckpt_root=ckpt_root,
                  val_data=val_data,
                  val_step=cfg.train.val_error_step,
                  checkpoint_step=cfg.train.checkpoint_step,
                  patience=cfg.train.patience,
                  single_ckpt=cfg.train.save_only_single_checkpoint_file,
                  device=cfg.train.device)



"""utility functions for working with audio formats"""
from evfuncs import load_cbin
from scipy.io import wavfile

from ..util.general import _files_from_dir


def swap_return_tuple_elements(func):
    def new_f(*args, **kwargs):
        return_tuple = func(*args, **kwargs)
        return return_tuple[1], return_tuple[0]
    return new_f


load_cbin = swap_return_tuple_elements(load_cbin)


AUDIO_FORMAT_FUNC_MAP = {
    'cbin': load_cbin,
    'wav': wavfile.read
}


def files_from_dir(audio_dir, audio_format):
    """get all audio files of a given format
    from a directory or its sub-directories,
    using the file extension associated with that annotation format.

    Parameters
    ----------
    audio_dir : str
        path to directory containing audio files.
    audio_format : str
        valid audio file format. One of {'wav', 'cbin'}.

    Returns
    -------
    audio_files : list
        of paths to audio files
    """
    if audio_format not in AUDIO_FORMAT_FUNC_MAP:
        raise ValueError(f"'{audio_format}' is not a valid audio format")
    audio_files = _files_from_dir(audio_dir, audio_format)
    return audio_files



import warnings

from crowsetta import Transcriber
import numpy as np

from .utils import _validate_durs
from .splitalgos import brute_force


def train_test_dur_split_inds(durs,
                              labels,
                              labelset,
                              train_dur,
                              test_dur,
                              val_dur=None,
                              algo='brute_force'):
    """return indices to split a dataset into training, test, and validation sets of specified durations.

    Given the durations of a set of vocalizations, and labels from the annotations for those vocalizations,
    this function returns arrays of indices for splitting up the set into training, test,
    and validation sets.

    Using those indices will produce datasets that each contain instances of all labels in the set of labels.

    Parameters
    ----------
    durs : iterable
        of float. Durations of audio files.
    labels : iterable
        of numpy arrays of str or int. Labels for segments (syllables, phonemes, etc.) in audio files.
    labelset : set, list
        set of unique labels for segments in files. Used to verify that each returned array
        of indices will produce a set that contains instances of all labels found in original
        set.
    train_dur : float
        Target duration for training set, in seconds.
    test_dur : float
        Target duration for test set, in seconds.
    val_dur : float
        Target duration for validation set, in seconds. Default is None.
        If None, no indices are returned for validation set.
    algo : str
        algorithm to use. One of {'brute_force', 'inc_freq'}. Default is 'brute_force'. For more information
        on the algorithms, see the docstrings, e.g., vak.io.splitalgos.brute_force
.
    Returns
    -------
    train_inds, test_inds, val_inds : numpy.ndarray
        indices to use with some array-like object to produce sets of specified durations
    """
    if len(durs) != len(labels):
        raise ValueError(
            f"length of durs, {len(durs)} does not equal length of labels, {len(labels)}"
        )

    total_dur = sum(durs)
    train_dur, val_dur, test_dur = _validate_durs(train_dur, val_dur, test_dur, total_dur)

    if -1 not in (train_dur, val_dur, test_dur):
        total_target_dur = sum([dur for dur in (train_dur, test_dur, val_dur) if dur is not None])

        if total_target_dur < total_dur:
            warnings.warn(
                'Total target duration of training, test, and (if specified) validation sets, '
                f'{total_target_dur} seconds, is less than total duration of dataset: {total_dur:.3f}. '
                'Not all of dataset will be used.'
            )

        if total_target_dur > total_dur:
            raise ValueError(
                f'Total duration of dataset, {total_dur} seconds, is less than total target duration of '
                f'training, test, and (if specified) validation sets: {total_target_dur}'
            )

    if algo == 'brute_force':
        train_inds,  val_inds, test_inds = brute_force(durs,
                                                       labels,
                                                       labelset,
                                                       train_dur,
                                                       val_dur,
                                                       test_dur)
    else:
        raise NotImplementedError(
            f'algorithm {algo} not implemented'
        )

    return train_inds, val_inds, test_inds


def train_test_dur_split(vak_df,
                         labelset,
                         train_dur=None,
                         test_dur=None,
                         val_dur=None):
    """split a dataset of vocalizations into training, test, and (optionally) validation subsets,
    specified by their duration.

    Takes dataset represented as a pandas DataFrame and adds a 'split' column that assigns each
    row to 'train', 'val', 'test', or 'None'.

    Parameters
    ----------
    vak_df : pandas.Dataframe
        a dataset of vocalizations.
    labelset : set, list
        of str or int, set of labels for vocalizations.
    train_dur : float
        total duration of training set, in seconds. Default is None
    test_dur : float
        total duration of test set, in seconds. Default is None.
    val_dur : float
        total duration of validation set, in seconds. Default is None.

    Returns
    -------
    vak_df : pandas.Dataframe
        with 'split' column added, that assigns each vocalization (row) to a subset, i.e., train, validation, or test.
        If the vocalization was not added to one of the subsets, its value for 'split' will be 'None'.

    Notes
    -----
    uses the function `vak.dataset.split.train_test_dur_split_inds` to find indices for each subset.
    """
    annot_format = vak_df['annot_format'].unique()
    if len(annot_format) == 1:
        annot_format = annot_format.item()
        # if annot_format is None, throw an error -- otherwise continue on and try to use it
        if annot_format is None:
            raise ValueError(
                'unable to load labels for dataset, the annot_format is None'
            )
    elif len(annot_format) > 1:
        raise ValueError(
            f'unable to load labels for dataset, found multiple annotation formats: {annot_format}'
        )
    # TODO: change this keyword argument to annot_format when changing dependency to crowsetta 2.0
    scribe = Transcriber(annot_format=annot_format)
    labels = [scribe.from_file(annot_file=annot_file).seq.labels
              for annot_file in vak_df['annot_path'].values]

    durs = vak_df['duration'].values
    total_dataset_dur = durs.sum()
    train_dur, val_dur, test_dur = _validate_durs(train_dur, val_dur, test_dur, total_dataset_dur)

    train_inds, val_inds, test_inds = train_test_dur_split_inds(durs=durs,
                                                                labels=labels,
                                                                labelset=labelset,
                                                                train_dur=train_dur,
                                                                test_dur=test_dur,
                                                                val_dur=val_dur)

    # start off with all elements set to 'None'
    # so we don't have to change any that are not assigned to one of the subsets to 'None' after
    split_col = np.asarray(['None' for _ in range(len(vak_df))], dtype='object')
    split_zip = zip(
        ['train', 'val', 'test'],
        [train_inds, val_inds, test_inds]
    )
    for split_name, split_inds in split_zip:
        if split_inds is not None:
            split_col[split_inds] = split_name

    # add split column to dataframe
    vak_df['split'] = split_col

    return vak_df



from .__about__ import (
    __author__,
    __commit__,
    __copyright__,
    __email__,
    __license__,
    __summary__,
    __title__,
    __uri__,
    __version__,
)

from . import __main__
from . import cli
from . import config
from . import datasets
from . import engine
from . import io
from . import metrics
from . import models
from . import transforms
from . import util

from .engine.model import Model


__all__ = [
    '__main__',
    'cli',
    'config',
    'datasets',
    'engine',
    'io',
    'metrics',
    'Model',
    'models',
    'transforms',
    'util'
]



import numpy as np

from ..util.path import array_dict_from_path
from ..util.validation import column_or_1d

from . import functional as F

__all__ = [
    'StandardizeSpect',
]


# adapted from:
# https://github.com/NickleDave/hybrid-vocal-classifier/blob/master/hvc/neuralnet/utils.py
class StandardizeSpect:
    """transform that standardizes spectrograms so they are all
    on the same scale, by subtracting off the mean and dividing by the
    standard deviation from a 'fit' set of spectrograms.

    Attributes
    ----------
    mean_freqs : numpy.ndarray
        mean values for each frequency bin across the fit set of spectrograms
    std_freqs : numpy.ndarray
        standard deciation for each frequency bin across the fit set of spectrograms
    non_zero_std : numpy.ndarray
        boolean, indicates where std_freqs has non-zero values. Used to avoid divide-by-zero errors.
    """
    def __init__(self,
                 mean_freqs=None,
                 std_freqs=None,
                 non_zero_std=None):
        """initialize a new StandardizeSpect instance

        Parameters
        ----------
        mean_freqs : numpy.ndarray
            vector of mean values for each frequency bin across the fit set of spectrograms
        std_freqs : numpy.ndarray
            vector of standard deviations for each frequency bin across the fit set of spectrograms
        non_zero_std : numpy.ndarray
            boolean, indicates where std_freqs has non-zero values. Used to avoid divide-by-zero errors.
        """
        if any([arg is not None for arg in (mean_freqs, std_freqs, non_zero_std)]):
            mean_freqs, std_freqs, non_zero_std = (
                column_or_1d(arr) for arr in (mean_freqs, std_freqs, non_zero_std)
            )
            if len(np.unique([arg.shape[0] for arg in (mean_freqs, std_freqs, non_zero_std)])) != 1:
                raise ValueError(
                    f'mean_freqs, std_freqs, and non_zero_std must all have the same length'
                )

        self.mean_freqs = mean_freqs
        self.std_freqs = std_freqs
        self.non_zero_std = non_zero_std

    @classmethod
    def fit_df(cls, df, spect_key='s'):
        """fits StandardizeSpect instance, given a pandas DataFrame representing a dataset

        Parameters
        ----------
        df : pandas.DataFrame
            loaded from a .csv file representing a dataset, created by vak.io.dataframe.from_files
        spect_key : str
            key in files in 'spect_path' column that maps to spectrograms in arrays.
            Default is 's'.

        Returns
        -------
        standardize_spect : StandardizeSpect
            instance fit to spectrograms in df
        """
        spect_paths = df['spect_path']
        spect = array_dict_from_path(spect_paths[0])[spect_key]
        # in files, spectrograms are in orientation (freq bins, time bins)
        # so we take mean and std across columns, i.e. time bins, i.e. axis 1
        mean_freqs = np.mean(spect, axis=1)
        std_freqs = np.std(spect, axis=1)

        for spect_path in spect_paths[1:]:
            spect = array_dict_from_path(spect_path)[spect_key]
            mean_freqs += np.mean(spect, axis=1)
            std_freqs += np.std(spect, axis=1)
        mean_freqs = mean_freqs / len(spect_paths)
        std_freqs = std_freqs / len(spect_paths)
        non_zero_std = np.argwhere(std_freqs != 0)
        return cls(mean_freqs, std_freqs, non_zero_std)

    @classmethod
    def fit(cls, spect):
        """fit a StandardizeSpect instance.

        Parameters
        ----------
        spect : numpy.ndarray
            with dimensions (frequency bins, time bins)

        Notes
        -----
        Input should be spectrogram.
        Fit function finds the mean and standard deviation of each frequency bin,
        which are used by `transform` method to scale other spectrograms.
        """
        # TODO: make this function accept list and/or ndarray with batch dimension
        if spect.ndim != 2:
            raise ValueError('input spectrogram should be a 2-d array')

        mean_freqs = np.mean(spect, axis=1)
        std_freqs = np.std(spect, axis=1)
        assert mean_freqs.shape[-1] == spect.shape[0]
        assert std_freqs.shape[-1] == spect.shape[0]
        non_zero_std = np.argwhere(std_freqs != 0)
        return cls(mean_freqs, std_freqs, non_zero_std)

    def __call__(self, spect):
        """normalizes input spectrogram with fit parameters.

        Parameters
        ----------
        spect : numpy.ndarray
            2-d array with dimensions (frequency bins, time bins).

        Returns
        -------
        z_norm_spect : numpy.ndarray
            array standardized to same scale as set of spectrograms that
            SpectScaler was fit with
        """
        if any([not hasattr(self, attr) for attr in ['mean_freqs',
                                                     'std_freqs']]):
            raise AttributeError('SpectScaler properties are set to None,'
                                 'must call fit method first to set the'
                                 'value of these properties before calling'
                                 'transform')

        if type(spect) != np.ndarray:
            raise TypeError(
                f'type of spect must be numpy.ndarray but was: {type(spect)}'
            )

        if spect.shape[0] != self.mean_freqs.shape[0]:
            raise ValueError(f'number of rows in spects, {spect.shape[0]}, '
                             f'does not match number of elements in self.mean_freqs, {self.mean_freqs.shape[0]},'
                             'i.e. the number of frequency bins from the spectrogram'
                             'to which the scaler was fit originally')

        return F.standardize_spect(spect, self.mean_freqs, self.std_freqs, self.non_zero_std)



"""parses [LEARNCURVE] section of config"""
import os
from configparser import NoOptionError

import attr
from attr.validators import instance_of

from .train import TrainConfig
from .. import models


@attr.s
class LearncurveConfig(TrainConfig):
    """class that represents [LEARNCURVE] section of config.ini file

    Attributes
    ----------
    models : list
        of model names. e.g., 'models = TweetyNet, GRUNet, ConvNet'
    csv_path : str
        path to where dataset was saved as a csv.
    num_epochs : int
        number of training epochs. One epoch = one iteration through the entire
        training set.
    normalize_spectrograms : bool
        if True, use spect.utils.data.SpectScaler to normalize the spectrograms.
        Normalization is done by subtracting off the mean for each frequency bin
        of the training set and then dividing by the std for that frequency bin.
        This same normalization is then applied to validation + test data.
    checkpoint_step : int
        step/epoch at which to save to checkpoint file.
        Default is None, in which case checkpoint is only saved at the last epoch.
    patience : int
        number of epochs to wait without the error dropping before stopping the
        training. Default is None, in which case training continues for num_epochs
    train_set_durs : list
        of int, durations in seconds of subsets taken from training data
        to create a learning curve, e.g. [5, 10, 15, 20]. Default is None
        (when training a single model on all available training data).
    num_replicates : int
        number of times to replicate training for each training set duration
        to better estimate mean accuracy for a training set of that size.
        Each replicate uses a different randomly drawn subset of the training
        data (but of the same duration).
    save_only_single_checkpoint_file : bool
        if True, save only one checkpoint file instead of separate files every time
        we save. Default is True.
    use_train_subsets_from_previous_run : bool
        if True, use training subsets saved in a previous run. Default is False.
        Requires setting previous_run_path option in config.ini file.
    previous_run_path : str
        path to results directory from a previous run.
        Used for training if use_train_subsets_from_previous_run is True.
    """
    train_set_durs = attr.ib(validator=instance_of(list), kw_only=True)
    num_replicates = attr.ib(validator=instance_of(int), kw_only=True)


def parse_learncurve_config(config, config_path):
    """parse [LEARNCURVE] section of config.ini file

    Parameters
    ----------
    config : ConfigParser
        containing config.ini file already loaded by parse function
    config_path : str
        path to config.ini file (used for error messages)

    Returns
    -------
    learncurve_config : vak.config.train.LearncurveConfig
        instance of LearncurveConfig class
    """
    config_dict = {}

    # load entry points within function, not at module level,
    # to avoid circular dependencies
    MODEL_NAMES = [model_name for model_name, model_builder in models.find()]
    try:
        model_names = [model_name
                       for model_name in config['LEARNCURVE']['models'].split(',')]
    except NoOptionError:
        raise KeyError("'models' option not found in [LEARNCURVE] section of config.ini file. "
                       "Please add this option as a comma-separated list of model names, e.g.:\n"
                       "models = TweetyNet, GRUnet, convnet")
    for model_name in model_names:
        if model_name not in MODEL_NAMES:
            raise ValueError(
                f'Model {model_name} not found when importing installed models.'
            )
    config_dict['models'] = model_names

    try:
        config_dict['train_vds_path'] = os.path.expanduser(config['LEARNCURVE']['train_vds_path'])
    except NoOptionError:
        raise KeyError("'train_vds_path' option not found in [LEARNCURVE] section of config.ini file. "
                       "Please add this option.")

    try:
        root_results_dir = config['LEARNCURVE']['root_results_dir']
        config_dict['root_results_dir'] = os.path.expanduser(root_results_dir)
    except NoOptionError:
        raise KeyError('must specify root_results_dir in [LEARNCURVE] section '
                       'of config.ini file')

    if config.has_option('LEARNCURVE', 'results_dir_made_by_main_script'):
        # don't check whether it exists because it may not yet,
        # depending on which function we are calling.
        # So it's up to calling function to check for existence of directory
        results_dirname = config['LEARNCURVE']['results_dir_made_by_main_script']
        config_dict['results_dirname'] = os.path.expanduser(results_dirname)
    else:
        config_dict['results_dirname'] = None

    if config.has_option('LEARNCURVE', 'train_set_durs'):
        config_dict['train_set_durs'] = [int(element)
                                         for element in
                                         config['LEARNCURVE']['train_set_durs'].split(',')]

    if config.has_option('LEARNCURVE', 'replicates'):
        config_dict['num_replicates'] = int(config['LEARNCURVE']['replicates'])

    if config.has_option('LEARNCURVE', 'val_vds_path'):
        config_dict['val_vds_path'] = os.path.expanduser(config['LEARNCURVE']['val_vds_path'])

    if config.has_option('LEARNCURVE', 'test_vds_path'):
        config_dict['test_vds_path'] = os.path.expanduser(config['LEARNCURVE']['test_vds_path'])

    if config.has_option('LEARNCURVE', 'val_error_step'):
        config_dict['val_error_step'] = int(config['LEARNCURVE']['val_error_step'])

    if config.has_option('LEARNCURVE', 'checkpoint_step'):
        config_dict['checkpoint_step'] = int(config['LEARNCURVE']['checkpoint_step'])

    if config.has_option('LEARNCURVE', 'save_only_single_checkpoint_file'):
        config_dict['save_only_single_checkpoint_file'] = config.getboolean(
            'LEARNCURVE', 'save_only_single_checkpoint_file'
        )

    if config.has_option('LEARNCURVE', 'num_epochs'):
        config_dict['num_epochs'] = int(config['LEARNCURVE']['num_epochs'])

    if config.has_option('LEARNCURVE', 'patience'):
        patience = config['LEARNCURVE']['patience']
        try:
            patience = int(patience)
        except ValueError:
            if patience == 'None':
                patience = None
            else:
                raise TypeError('patience must be an int or None, but'
                                'is {} and parsed as type {}'
                                .format(patience, type(patience)))
        config_dict['patience'] = patience

    if config.has_option('LEARNCURVE', 'normalize_spectrograms'):
        config_dict['normalize_spectrograms'] = config.getboolean(
            'LEARNCURVE', 'normalize_spectrograms'
        )

    if config.has_option('LEARNCURVE', 'use_train_subsets_from_previous_run'):
        config_dict['use_train_subsets_from_previous_run'] = config.getboolean(
            'LEARNCURVE', 'use_train_subsets_from_previous_run')
        if config_dict['use_train_subsets_from_previous_run']:
            try:
                config_dict['previous_run_path'] = os.path.expanduser(config['LEARNCURVE']['previous_run_path'])
            except KeyError:
                raise KeyError('In config.file {}, '
                               'use_train_subsets_from_previous_run = Yes, but '
                               'no previous_run_path option was found.'
                               'Please add previous_run_path to config file.'
                               .format(config_path))
        else:
            if config.has_option('LEARNCURVE', 'previous_run_path'):
                raise ValueError('In config.file {}, '
                                 'use_train_subsets_from_previous_run = No, but '
                                 'previous_run_path option was specified as {}.\n'
                                 'Please fix argument or remove/comment out '
                                 'previous_run_path.'
                                 .format(config_path,
                                         config['LEARNCURVE']['previous_run_path'])
                                 )

    if config.has_option('LEARNCURVE', 'save_transformed_data'):
        config_dict['save_transformed_data'] = config.getboolean(
            'LEARNCURVE', 'save_transformed_data')

    return LearncurveConfig(**config_dict)



import random
import itertools
import logging

from ..utils import _validate_durs


def brute_force(durs,
                labels,
                labelset,
                train_dur,
                val_dur,
                test_dur,
                max_iter=5000):
    """finds indices that split (labels, durations) tuples into training,
    test, and validation sets of specified durations, with the set of unique labels
    in each dataset equal to the specified labelset.

    The durations of the datasets created using the returned indices will be *greater than* or equal to
    the durations specified.

    Must specify a positive value for one of {train_dur, test_dur}.
    The other value can be specified as '-1' which is interpreted as "use the
    remainder of the dataset for this split after finding indices for the set with a specified duration".
    If only one of {train_dur, test_dur} is specified, the other defaults to '-1'.

    Parameters
    ----------
    durs : list
        of durations of vocalizations
    labels : list
        of labels from vocalizations
    labelset : set
        of labels
    train_dur : int, float
        Target duration for training set, in seconds.
    val_dur : int, float
        Target duration for validation set, in seconds.
    test_dur : int, float
        Target duration for test set, in seconds.
    max_iter : int
        maximum number of iterations to attempt to find indices. Default is 5000.

    Returns
    -------
    train_inds, val_inds, test_inds : list
        of int, the indices that will split datasets

    Notes
    -----
    A 'brute force' algorithm that just randomly assigns indices to a set,
    and iterates until it finds some partition where each set has instances of all classes of label.
    """
    logger = logging.getLogger(__name__)
    logger.setLevel('INFO')

    sum_durs = sum(durs)
    train_dur, val_dur, test_dur = _validate_durs(train_dur, val_dur, test_dur, sum_durs)

    if -1 not in (train_dur, test_dur):
        total_target_dur = sum([dur for dur in (train_dur, val_dur, test_dur) if dur is not None])
    else:
        total_target_dur = sum_durs

    durs_labels_list = list(zip(durs, labels))
    iter = 1
    all_labels_err = ('Did not successfully divide data into training, '
                      'validation, and test sets of sufficient duration '
                      f'after {max_iter} iterations.'
                      ' Try increasing the total size of the data set.')

    while 1:
        train_inds = []
        test_inds = []
        if val_dur:
            val_inds = []
        else:
            val_inds = None

        total_train_dur = 0
        total_val_dur = 0
        total_test_dur = 0

        durs_labels_inds = list(range(len(durs_labels_list)))
        random.shuffle(durs_labels_inds)

        finished = False

        if -1 in (train_dur, test_dur):
            if train_dur == -1:
                choice = ['test', 'train']
            elif test_dur == -1:
                choice = ['train', 'test']

            while finished is False:
                try:
                    ind = durs_labels_inds.pop()
                except IndexError:
                    if len(durs_labels_inds) == 0:
                        logger.debug(
                            'Ran out of elements while dividing dataset into subsets of specified durations.'
                            f'Iteration {iter}'
                        )
                        iter += 1
                        break  # do next iteration
                    else:
                        # something else happened, re-raise error
                        raise
                which_set = choice[0]
                if which_set == 'train':
                    train_inds.append(ind)
                    total_train_dur += durs_labels_list[ind][0]  # ind 0 is duration
                    if train_dur != -1 and total_train_dur >= train_dur:
                        choice.pop(choice.index('train'))
                elif which_set == 'test':
                    test_inds.append(ind)
                    total_test_dur += durs_labels_list[ind][0]  # ind 0 is duration
                    if test_dur != -1 and total_test_dur >= test_dur:
                        choice.pop(choice.index('test'))

                if len(durs_labels_inds) == 0:
                    finished = True
                    break

        else:
            choice = []
            if train_dur > 0:
                choice.append('train')
            if test_dur > 0:
                choice.append('test')
            if val_dur:
                choice.append('val')

            while finished is False:
                # pop durations off list and append to randomly-chosen
                # list, either train, val, or test set.
                # Do this until the total duration for each data set is equal
                # to or greater than the target duration for each set.
                try:
                    ind = durs_labels_inds.pop()
                except IndexError:
                    if len(durs_labels_inds) == 0:
                        logger.debug(
                            'Ran out of elements while dividing dataset into subsets of specified durations.'
                            f'Iteration {iter}'
                        )
                        iter += 1
                        break  # do next iteration
                    else:
                        # something else happened, re-raise error
                        raise

                which_set = random.randint(0, len(choice) - 1)
                which_set = choice[which_set]
                if which_set == 'train':
                    train_inds.append(ind)
                    total_train_dur += durs_labels_list[ind][0]  # ind 0 is duration
                    if total_train_dur >= train_dur:
                        choice.pop(choice.index('train'))
                elif which_set == 'val':
                    val_inds.append(ind)
                    total_val_dur += durs_labels_list[ind][0]  # ind 0 is duration
                    if total_val_dur >= val_dur:
                        choice.pop(choice.index('val'))
                elif which_set == 'test':
                    test_inds.append(ind)
                    total_test_dur += durs_labels_list[ind][0]  # ind 0 is duration
                    if total_test_dur >= test_dur:
                        choice.pop(choice.index('test'))

                if len(choice) < 1:
                    total_all_durs = sum([dur for dur in (total_train_dur, total_val_dur, total_test_dur)
                                          if dur is not None])
                    if total_all_durs < total_target_dur:
                        raise ValueError(
                            f'Loop to find subsets completed but total duration of subsets, {total_all_durs} seconds, '
                            f'is less than total duration specified: {total_target_dur} seconds.')
                    else:
                        finished = True
                        break

            if iter > max_iter:
                raise ValueError('Could not find subsets of sufficient duration in '
                                 f'less than {max_iter} iterations.')

        if finished is True:
            # make sure that each set contains all classes we
            # want the network to learn
            train_tups = [durs_labels_list[ind] for ind in train_inds]  # tup = a tuple
            test_tups = [durs_labels_list[ind] for ind in test_inds]
            if val_dur:
                val_tups = [durs_labels_list[ind] for ind in val_inds]

            train_labels = itertools.chain.from_iterable(
                [tup[1] for tup in train_tups])
            train_labels = set(train_labels)  # make set to get unique values

            if test_dur > 0 or test_dur == -1:
                test_labels = itertools.chain.from_iterable(
                    [tup[1] for tup in test_tups])
                test_labels = set(test_labels)

            if val_dur:
                val_labels = itertools.chain.from_iterable(
                    [tup[1] for tup in val_tups])
                val_labels = set(val_labels)

            if train_labels != set(labelset):
                iter += 1
                if iter > max_iter:
                    raise ValueError(all_labels_err)
                else:
                    logger.debug(
                        'Train labels did not contain all labels in labelset. '
                        f'Getting new training set. Iteration {iter}'
                    )
                    continue

            if test_dur > 0 or test_dur == -1:
                if test_labels != set(labelset):
                    iter += 1
                    if iter > max_iter:
                        raise ValueError(all_labels_err)
                    else:
                        logger.debug(
                            'Test labels did not contain all labels in labelset. '
                            f'Getting new test set. Iteration {iter}'
                        )
                        continue

            if val_dur is not None and val_labels != set(labelset):
                iter += 1
                if iter > max_iter:
                    raise ValueError(all_labels_err)
                else:
                    logger.debug(
                        'Validation labels did not contain all labels in labelset. '
                        f'Getting new validation set. Iteration {iter}'
                    )
                    continue
            else:
                break
        elif finished is False:
            continue

    return train_inds, val_inds, test_inds



import crowsetta
import numpy as np

from .validation import column_or_1d


def has_unlabeled(labels_int,
                  onsets_s,
                  offsets_s,
                  time_bins):
    """determine whether there are unlabeled segments in a spectrogram,
    given labels, onsets, and offsets of vocalizations, and vector of
    time bins from spectrogram

    Parameters
    ----------
    labels_int : list, numpy.ndarray
        a list or array of labels from the annotation for a vocalization,
        mapped to integers
    onsets_s : numpy.ndarray
        1d vector of floats, segment onsets in seconds
    offsets_s : numpy.ndarray
        1-d vector of floats, segment offsets in seconds
    time_bins : numpy.ndarray
        1-d vector of floats, time in seconds for center of each time bin of a spectrogram

    Returns
    -------
    has_unlabeled : bool
        if True, there are time bins that do not have labels associated with them
    """
    if (type(labels_int) == list and not all([type(lbl) == int for lbl in labels_int]) or
            (type(labels_int) == np.ndarray and labels_int.dtype not in [np.int8, np.int16, np.int32, np.int64])):
        raise TypeError('labels_int must be a list or numpy.ndarray of integers')

    dummy_unlabeled_label = np.max(labels_int) + 1
    label_vec = np.ones((time_bins.shape[-1], 1), dtype='int8') * dummy_unlabeled_label
    onset_inds = [np.argmin(np.abs(time_bins - onset))
                  for onset in onsets_s]
    offset_inds = [np.argmin(np.abs(time_bins - offset))
                   for offset in offsets_s]
    for label, onset, offset in zip(labels_int, onset_inds, offset_inds):
        # offset_inds[ind]+1 because offset time bin is still "part of" syllable
        label_vec[onset:offset+1] = label

    if dummy_unlabeled_label in label_vec:
        return True
    else:
        return False


def label_timebins(labels_int,
                   onsets_s,
                   offsets_s,
                   time_bins,
                   unlabeled_label=0):
    """makes a vector of labels for each time bin from a spectrogram,
    given labels, onsets, and offsets of vocalizations

    Parameters
    ----------
    labels_int : list, numpy.ndarray
        a list or array of labels from the annotation for a vocalization,
        mapped to integers
    onsets_s : numpy.ndarray
        1d vector of floats, segment onsets in seconds
    offsets_s : numpy.ndarray
        1-d vector of floats, segment offsets in seconds
    time_bins : mumpy.ndarray
        1-d vector of floats, time in seconds for center of each time bin of a spectrogram
    unlabeled_label : int
        label assigned to time bins that do not have labels associated with them.
        Default is 0

    Returns
    -------
    lbl_tb : numpy.ndarray
        same length as time_bins, with each element a label for each time bin
    """
    if (type(labels_int) == list and not all([type(lbl) == int for lbl in labels_int]) or
            (type(labels_int) == np.ndarray and labels_int.dtype not in [np.int8, np.int16, np.int32, np.int64])):
        raise TypeError('labels_int must be a list or numpy.ndarray of integers')

    label_vec = np.ones((time_bins.shape[-1],), dtype='int8') * unlabeled_label
    onset_inds = [np.argmin(np.abs(time_bins - onset))
                  for onset in onsets_s]
    offset_inds = [np.argmin(np.abs(time_bins - offset))
                   for offset in offsets_s]
    for label, onset, offset in zip(labels_int, onset_inds, offset_inds):
        # offset_inds[ind]+1 because offset time bin is still "part of" syllable
        label_vec[onset:offset+1] = label

    return label_vec


def translate(label_arr, map_dict):
    return np.vectorize(map_dict.__getitem__)(label_arr)


def where(label_arrs, find_in_arr=False):
    """given a list of label arrays for a set of audio files,
    return for each unique label the indices of the arrays in which it occurs
    and optionally where it occurs in each array of labels.

    What is actually returned are index arrays for each unique label.
    These can be used for example to index into a list of file names that corresponds
    to the list of label arrays passed to this function.

    Parameters
    ----------
    label_arrs : list, tuple
        of arrays of labels for segments in a set of audio files, e.g. an array
        of labels for animal vocalizations and silent periods within each file.
        Each element in the iterable should correspond to a labels array for one file.
    find_in_arr : bool
        if True, find where each unique label occurs in each array of labels, i.e.,
        also return indices of each occurrence within all arrays.
        Default is False.

    Returns
    -------
    where_in_label_arrs : dict
        in which keys are unique labels, and the corresponding
        value for each key is an array of indices that would select
        all elements in labels that
    where_in_each_arr : dict
        in which keys are unique labels, and the corresponding
        value for each key is a dictionary where the keys are indices
        into label_arrs, and the values are index arrays that would
        select just the specified label from the array it is mapped to.
    """
    if type(label_arrs) not in [list, tuple]:
        raise TypeError(f"label_arrs must be a list or tuple, but was: {type(labels)}")

    label_arrs = [column_or_1d(lbl) for lbl in label_arrs]

    label_arrs_inds = []
    all_lbl_arr = []

    for ind, label_arr in enumerate(label_arrs):
        all_lbl_arr.append(label_arr)
        label_arrs_inds.append(
            np.ones(shape=label_arr.shape[0], dtype=np.int64) * ind
        )

    all_lbl_arr = np.concatenate(all_lbl_arr)
    label_arrs_inds = np.concatenate(label_arrs_inds)

    uniq_lbl = np.unique(all_lbl_arr)
    where_in_labels = {}

    for this_lbl in uniq_lbl:
        in_all_lbl = np.where(all_lbl_arr == this_lbl)[0]
        this_lbl_label_arrs_inds = label_arrs_inds[in_all_lbl]
        uniq_label_arrs_inds = np.unique(this_lbl_label_arrs_inds)
        where_in_labels[this_lbl] = uniq_label_arrs_inds

    if find_in_arr:
        where_in_arr = {}
        for this_lbl, these_inds in where_in_labels.items():
            ind_arr_map = {}
            for this_ind in these_inds:
                in_this_arr = np.where(label_arrs[this_ind] == this_lbl)[0]
                ind_arr_map[this_ind] = in_this_arr
            where_in_arr[this_lbl] = ind_arr_map

    if find_in_arr:
        return where_in_labels, where_in_arr
    else:
        return where_in_labels


def sort(uniq_labels, occurrences, ascending=True):
    """sort a list of unique labels by the number of times they occur.

    Parameters
    ----------
    uniq_labels : list
        list of unique label classes
    occurrences : list
        of index arrays, same length as uniq_labels.
        Each array in the list contains indexes where the corresponding label
        in uniq_labels occurs, e.g. in a set of annotations for audio files, or
        in an array of labels for an individual audio file.
    ascending : bool
        if True, sort in ascending order. if False, return in descending order.
        Default is True.

    Returns
    -------
    uniq_labels_sorted, occurrences_sorted : list
    """
    if len(set(uniq_labels)) != len(uniq_labels):
        raise ValueError('uniq_labels should be a unique set, but found repeated elements')

    occurrences = [column_or_1d(occur_arr) for occur_arr in occurrences]

    counts = [arr.shape[0] for arr in occurrences]
    sort_inds = np.argsort(counts)
    # make lists since occurrences wil have different lengths for each array
    uniq_labels_sorted = []
    occurrences_sorted = []
    for sort_ind in sort_inds:
        uniq_labels_sorted.append(uniq_labels[sort_ind])
        occurrences_sorted.append(occurrences[sort_ind])
    if ascending:
        return uniq_labels_sorted, occurrences_sorted
    else:
        return uniq_labels_sorted.reverse(), occurrences_sorted.reverse()


def to_map(labelset, map_unlabeled=True):
    """map set of labels to series of consecutive integers from 0 to n inclusive,
    where n is the number of labels in the set.

    This 'labelmap' is used when mapping labels from annotations of a vocalization into
    a label for every time bin in a spectrogram of that vocalization.

    If map_unlabeled is True, 'unlabeled' will be added to labelset, and will map to 0,
    so the total number of classes is n + 1.

    Parameters
    ----------
    labelset : set
        of labels used to annotate a Dataset.
    map_unlabeled : bool
        if True, include key 'unlabeled' in mapping. Any time bins in a spectrogram
        that do not have a label associated with them, e.g. a silent gap between vocalizations,
        will be assigned the integer that the 'unlabeled' key maps to.

    Returns
    -------
    labelmap : dict
        maps labels to integers
    """
    if type(labelset) != set:
        raise TypeError(
            f'type of labelset must be set, got type {type(labelset)}'
        )

    labellist = []
    if map_unlabeled is True:
        labellist.append('unlabeled')

    labellist.extend(
        sorted(list(labelset))
    )

    labelmap = dict(
        zip(
            labellist, range(len(labellist))
        )
    )
    return labelmap


def to_set(labels_list):
    """given a list of labels from annotations, return the set of (unique) labels

    Parameters
    ----------
    labels_list : list
         of lists, i.e. labels from annotations

    Returns
    -------
    labelset

    Examples
    --------
    >>> labels_list = [voc.annot.labels for voc in vds.voc_list]
    >>> labelset = to_set(labels_list)
    """
    all_labels = [lbl for labels in labels_list for lbl in labels]
    labelset = set(all_labels)
    return labelset


def lbl_tb2labels(labeled_timebins,
                  labels_mapping,
                  spect_ID_vector=None):
    """converts output of network from label for each frame
    to one label for each continuous segment

    Parameters
    ----------
    labeled_timebins : ndarray
        where each element is a label for a time bin.
        Such an array is the output of the network.
    labels_mapping : dict
        that maps str labels to consecutive integers.
        The mapping is inverted to convert back to str labels.
    spect_ID_vector : ndarray
        of same length as labeled_timebins, where each element
        is an ID # for the spectrogram from which labeled_timebins
        was taken.
        If provided, used to split the converted labels back to
        a list of label str, with one for each spectrogram.
        Default is None, in which case the return value is one long str.

    Returns
    -------
    labels : str or list
        labeled_timebins mapped back to label str.
        If spect_ID_vector was provided, then labels is split into a list of str,
        where each str corresponds to predicted labels for each predicted
        segment in each spectrogram as identified by spect_ID_vector.
    """
    idx = np.diff(labeled_timebins, axis=0).astype(np.bool)
    idx = np.insert(idx, 0, True)

    labels = labeled_timebins[idx]

    # remove 'unlabeled' label
    if 'unlabeled' in labels_mapping:
        labels = labels[labels != labels_mapping['unlabeled']]
        labels = labels.tolist()

    inverse_labels_mapping = dict((v, k) for k, v
                                  in labels_mapping.items())
    labels = [inverse_labels_mapping[label] for label in labels]

    if spect_ID_vector:
        labels_list = []
        spect_ID_vector = spect_ID_vector[idx]
        labels_arr = np.asarray(labels)
        # need to split up labels by spect_ID_vector
        # this is probably not the most efficient way:
        spect_IDs = np.unique(spect_ID_vector)

        for spect_ID in spect_IDs:
            these = np.where(spect_ID_vector == spect_ID)
            curr_labels = labels_arr[these].tolist()
            if all([type(el) is str for el in curr_labels]):
                labels_list.append(''.join(curr_labels))
            elif all([type(el) is int for el in curr_labels]):
                labels_list.append(curr_labels)
        return labels_list, spect_ID_vector
    else:
        if all([type(el) is str or type(el) is np.str_ for el in labels]):
            return ''.join(labels)
        elif all([type(el) is int for el in labels]):
            return labels


def _segment_lbl_tb(lbl_tb):
    """helper function that segments vector of labeled timebins.

    Parameters
    ----------
    lbl_tb : numpy.ndarray
        vector where each element represents a label for a timebin

    Returns
    -------
    labels : numpy.ndarray
        vector where each element is a label for a segment with its onset
        and offset indices given by the corresponding element in onset_inds
        and offset_inds.
    onset_inds : numpy.ndarray
        vector where each element is the onset index for a segment.
        Each onset corresponds to the value at the same index in labels.
    offset_inds : numpy.ndarray
        vector where each element is the offset index for a segment
        Each offset corresponds to the value at the same index in labels.
    """
    # factored out as a separate function to be able to test
    # and in case user wants to do just this with output of neural net
    offset_inds = np.where(np.diff(lbl_tb, axis=0))[0]
    onset_inds = offset_inds + 1
    offset_inds = np.concatenate(
        (offset_inds, np.asarray([lbl_tb.shape[0] - 1]))
    )
    onset_inds = np.concatenate(
        (np.asarray([0]), onset_inds)
    )
    labels = lbl_tb[onset_inds]
    return labels, onset_inds, offset_inds


def lbl_tb2segments(lbl_tb,
                    labelmap,
                    timebin_dur):
    """convert vector of labeled timebins into segments,
    by finding where continuous runs of a single label start
    and stop. Returns vectors of labels and onsets and offsets
    in units of seconds.

    Parameters
    ----------
    lbl_tb : numpy.ndarray
        vector of labeled spectrogram time bins, i.e.,
        where each element is a label for a time bin.
        Output of a neural network.
    labelmap : dict
        that maps labels to consecutive integers.
        The mapping is inverted to convert back to labels.
    timebin_dur : float
        Duration of a single timebin in the spectrogram, in seconds.
        Used to convert onset and offset indices in lbl_tb to seconds.

    Returns
    -------
    labels : numpy.ndarray
        vector where each element is a label for a segment with its onset
        and offset indices given by the corresponding element in onset_inds
        and offset_inds.
    onsets_s : numpy.ndarray
        vector where each element is the onset in seconds a segment.
        Each onset corresponds to the value at the same index in labels.
    offsets_s : numpy.ndarray
        vector where each element is the offset in seconds of a segment.
        Each offset corresponds to the value at the same index in labels.
    """
    lbl_tb = column_or_1d(lbl_tb)
    labels, onset_inds, offset_inds = _segment_lbl_tb(lbl_tb)

    # remove 'unlabeled' label
    if 'unlabeled' in labelmap:
        keep = np.where(labels != labelmap['unlabeled'])[0]
        labels = labels[keep]
        onset_inds = onset_inds[keep]
        offset_inds = offset_inds[keep]
    inverse_labelmap = dict((v, k) for k, v
                            in labelmap.items())
    labels = labels.tolist()
    labels = np.asarray(
        [inverse_labelmap[label] for label in labels]
    )
    onsets_s = onset_inds * timebin_dur
    offsets_s = offset_inds * timebin_dur

    return labels, onsets_s, offsets_s



"""parses [SPECT_PARAMS] section of config"""
import attr
from attr import converters, validators
from attr.validators import instance_of


def freq_cutoffs_validator(instance, attribute, value):
    if len(value) != 2:
        raise ValueError(
            f'freq_cutoffs should be a list of two elements, but instead got: {value}')
    if value[0] > value[1]:
        raise ValueError(
            f'lower freq_cutoff should be less than higher freq_cutoff, instead of: {value}'
        )


VALID_TRANSFORM_TYPES = {'log_spect', 'log_spect_plus_one'}


def is_valid_transform_type(instance, attribute, value):
    if value not in VALID_TRANSFORM_TYPES:
        raise ValueError(
            f'Value for `transform_type`, {value}, in [SPECTROGRAM] '
            'section of .ini file is not recognized. Must be one '
            f'of the following: {VALID_TRANSFORM_TYPES}'
        )


@attr.s
class SpectParamsConfig:
    """represents parameters for making spectrograms from audio and saving in files

    Attributes
    ----------
    fft_size : int
        size of window for Fast Fourier transform, number of time bins. Default is 512.
    step_size : int
        step size for Fast Fourier transform. Default is 64.
    freq_cutoffs : tuple
        of two elements, lower and higher frequencies. Used to bandpass filter audio
        (using a Butter filter) before generating spectrogram.
        Default is None, in which case no bandpass filtering is applied.
    transform_type : str
        one of {'log_spect', 'log_spect_plus_one'}.
        'log_spect' transforms the spectrogram to log(spectrogram), and
        'log_spect_plus_one' does the same thing but adds one to each element.
        Default is None. If None, no transform is applied.
    thresh: int
        threshold minimum power for log spectrogram.
    spect_key : str
        key for accessing spectrogram in files. Default is 's'.
    freqbins_key : str
        key for accessing vector of frequency bins in files. Default is 'f'.
    timebins_key : str
        key for accessing vector of time bins in files. Default is 't'.
    audio_path_key : str
        key for accessing path to source audio file for spectogram in files.
        Default is 'audio_path'.
    """
    fft_size = attr.ib(converter=int, validator=instance_of(int), default=512)
    step_size = attr.ib(converter=int, validator=instance_of(int), default=64)
    freq_cutoffs = attr.ib(validator=validators.optional(freq_cutoffs_validator),
                           default=None)
    thresh = attr.ib(converter=converters.optional(float),
                     validator=validators.optional(instance_of(float)),
                     default=None)
    transform_type = attr.ib(validator=validators.optional([instance_of(str), is_valid_transform_type]),
                             default=None)
    spect_key = attr.ib(validator=instance_of(str), default='s')
    freqbins_key = attr.ib(validator=instance_of(str), default='f')
    timebins_key = attr.ib(validator=instance_of(str), default='t')
    audio_path_key = attr.ib(validator=instance_of(str), default='audio_path')


def parse_spect_params_config(config_toml, toml_path):
    """parse [SPECT_PARAMS] section of config.toml file

    Parameters
    ----------
    config_toml : dict
        containing configuration file in TOML format, already loaded by parse function
    toml_path : Path
        path to a configuration file in TOML format (used for error messages)

    Returns
    -------
    spect_params_config : vak.config.spect_params.SpectParamsConfig
        instance with attributes set to values specified by config.toml section
        or to defaults.
    """
    # return defaults if config doesn't have SPECT_PARAMS section
    spect_params_section = {}
    if 'SPECT_PARAMS' in config_toml:
        spect_params_section.update(
            config_toml['SPECT_PARAMS'].items()
        )
    return SpectParamsConfig(**spect_params_section)



import os


def predict(predict_vds_path,
            train_vds_path,
            checkpoint_path,
            networks,
            spect_scaler_path=None,
            ):
    """make predictions with one trained model

    Parameters
    ----------
    predict_vds_path : str
        path to saved Dataset that contains data for which annotations
        should be predicted.
    train_vds_path : str
        path to Dataset that represents training data.
        To fetch labelmap used during training, to map labels used
        in annotation to a series of consecutive integers that become
        outputs of the neural network. Used here to convert
        back to labels used in annotation.
    checkpoint_path : str
        path to directory with checkpoint files saved by Tensorflow, to reload model
    networks : namedtuple
        where each field is the Config tuple for a neural network and the name
        of that field is the name of the class that represents the network.
    spect_scaler_path : str
        path to a saved SpectScaler object used to normalize spectrograms.
        If spectrograms were normalized and this is not provided, will give
        incorrect results.

    Returns
    -------
    None
    """
    if not os.path.isdir(checkpoint_path):
        raise FileNotFoundError('directory {}, specified as '
                                'checkpoint_path, is not found.'
                                .format(checkpoint_path))

    if spect_scaler_path:
        if not os.path.isfile(spect_scaler_path):
            raise FileNotFoundError(
                f'file for spect_scaler not found at path:\n{spect_scaler_path}'
            )

    if type(predict_vds_path) is str:
        predict_vds_path = [predict_vds_path]
    elif type(predict_vds_path) is list:
        pass
    else:
        raise TypeError(
            'predict_vds_path should be a string path or list '
            f'of string paths, but type was {type(predict_vds_path)}'
        )

    train_vds = Dataset.load(json_fname=train_vds_path)
    if train_vds.are_spects_loaded() is False:
        train_vds = train_vds.load_spects()
    labelmap = train_vds.labelmap
    del train_vds



from pathlib import Path

from .train import train
from .predict import predict
from .learncurve import learning_curve
from .prep import prep
from .. import config
from ..config import parse_learncurve_config, parse_predict_config, parse_prep_config


def cli(command, config_file):
    """command-line interface

    Parameters
    ----------
    command : string
        One of {'prep', 'train', 'predict', 'finetune', 'learncurve'}
    config_file : str, Path
        path to a config.toml file
    """
    # check config_file exists,
    # because if it doesn't ConfigParser will just return an "empty" instance w/out sections or options
    config_file = Path(config_file)
    if not config_file.is_file():
        raise FileNotFoundError(f'file not found, or not recognized as a file: {config_file}')

    if command == 'prep':
        prep(toml_path=config_file)

    elif command == 'train':
        train(toml_path=config_file)

    elif command == 'finetune':
        raise NotImplementedError

    # elif command == 'predict':
    #     predict_config = parse_predict_config(config_obj)
    #     models = config.models.from_config(config_obj, predict_config.models)
    #     predict(predict_vds_path=predict_config.predict_vds_path,
    #             train_vds_path=predict_config.train_vds_path,
    #             checkpoint_path=predict_config.checkpoint_path,
    #             networks=models,
    #             spect_scaler_path=predict_config.spect_scaler_path)
    #
    # elif command == 'learncurve':
    #     if config_obj.has_option('TRAIN', 'results_dir_made_by_main_script'):
    #         raise ValueError(
    #             f"config file {config_file} already has option 'results_dir_made_by_main_script' "
    #             "in [TRAIN] section. \nRunning learncurve will overwrite that option."
    #             "Please either remove the option from this file or make a copy of the config.ini "
    #             "file with a new name and remove it from that file.\n"
    #             f"Currently the option is set to: {config_obj['TRAIN']['results_dir_made_by_main_script']}"
    #         )
    #     learncurve_config = parse_learncurve_config(config_obj, config_file)
    #     models = config.models.from_config(config_obj, learncurve_config.models)
    #     prep_config = parse_prep_config(config_obj, config_file)
    #     if learncurve_config.train_vds_path is None:
    #         raise ValueError("must set 'train_vds_path' option in [LEARNCURVE] section of config.ini file "
    #                          "before running 'learncurve'")
    #     if learncurve_config.val_vds_path is None:
    #         raise ValueError("must set 'val_vds_path' option in [LEARNCURVE] section of config.ini file "
    #                          "before running 'learncurve'")
    #     learning_curve(train_vds_path=learncurve_config.train_vds_path,
    #                    val_vds_path=learncurve_config.val_vds_path,
    #                    test_vds_path=learncurve_config.test_vds_path,
    #                    train_dur=prep_config.train_dur,
    #                    train_set_durs=learncurve_config.train_set_durs,
    #                    num_replicates=learncurve_config.num_replicates,
    #                    num_epochs=learncurve_config.num_epochs,
    #                    config_file=config_file,
    #                    networks=models,
    #                    val_error_step=learncurve_config.val_error_step,
    #                    checkpoint_step=learncurve_config.checkpoint_step,
    #                    patience=learncurve_config.patience,
    #                    save_only_single_checkpoint_file=learncurve_config.save_only_single_checkpoint_file,
    #                    normalize_spectrograms=learncurve_config.normalize_spectrograms,
    #                    use_train_subsets_from_previous_run=learncurve_config.use_train_subsets_from_previous_run,
    #                    previous_run_path=learncurve_config.previous_run_path,
    #                    root_results_dir=learncurve_config.root_results_dir,
    #                    save_transformed_data=learncurve_config.save_transformed_data)



"""utility functions that deal with paths"""
from functools import partial
from pathlib import Path

import numpy as np
from scipy.io import loadmat

from ..config import validators
from .general import find_fname


def find_audio_fname(spect_path, audio_ext=None):
    """finds name of audio file in a path to a spectogram file,
    if one is present.

    Checks for any extension that is a valid audio file format
    and returns path up to and including that extension,
    i.e. with the spectrogram file extension removed.

    Parameters
    ----------
    spect_path : str
        path to a spectrogram file
    audio_ext : str
        extension associated with an audio file format, used to
        find audio file name in spect_path.
        Default is None. If None, search for any valid audio format
        (as defined by vak.config.validators.VALID_AUDIO_FORMATS)

    Returns
    -------
    audio_fname : str
        name of audio file found in spect_path
    """
    if audio_ext is None:
        audio_ext = validators.VALID_AUDIO_FORMATS
    elif type(audio_ext) is str:
        audio_ext = [audio_ext]
    else:
        raise TypeError(
            f'invalid type for audio_ext: {type(audio_ext)}'
        )

    audio_fnames = []
    for ext in audio_ext:
        audio_fnames.append(
            find_fname(spect_path, ext)
        )
    # remove Nones
    audio_fnames = [path for path in audio_fnames if path is not None]
    # keep just file name from spect path
    audio_fnames = [Path(path).name for path in audio_fnames]

    if len(audio_fnames) == 1:
        return audio_fnames[0]
    else:
        raise ValueError(
            f'unable to determine filename of audio file from: {spect_path}'
        )


SPECT_FORMAT_LOAD_FUNCTION_MAP = {
    'mat': partial(loadmat, squeeze_me=True),
    'npz': np.load,
}


def array_dict_from_path(spect_path, spect_format=None):
    """load spectrogram and related arrays from a file,
    return as an object that provides Python dictionary-like
    access

    Parameters
    ----------
    spect_path : str, Path
        to an array file.
    spect_format : str
        Valid formats are defined in vak.io.spect.SPECT_FORMAT_LOAD_FUNCTION_MAP.
        Default is None, in which case the extension of the file is used.

    Returns
    -------
    spect_dict : dict-like
        either a dictionary or dictionary-like object that provides access to arrays
        from the file via keys, e.g. spect_dict['s'] for the spectrogram.
        See docstring for vak.audio.to_spect for default keys for spectrogram
        array files that function creates.
    """
    spect_path = Path(spect_path)
    if spect_format is None:
        # "replace('.', '')", because suffix returns file extension with period included
        spect_format = spect_path.suffix.replace('.', '')
    spect_dict = SPECT_FORMAT_LOAD_FUNCTION_MAP[spect_format](spect_path)
    return spect_dict



import logging
import os
from pathlib import Path
from datetime import datetime

import toml

from .. import config
from ..io import dataframe
from ..util import train_test_dur_split


def prep(toml_path):
    """prepare datasets from vocalizations.
    Function called by command-line interface.

    Parameters
    ----------
    toml_path : str, Path
        path to a configuration file in TOML format.
        Used to rewrite file with options determined by this function and needed for other functions

    Returns
    -------
    None

    Notes
    -----
    Saves a .csv file representing the dataset generated from data_dir.
    If durations were specified for validation and test sets, then the .csv
    has a column representing which files belong to the training, test, and
    validation sets created from that Dataset.

    Datasets are used to train neural networks that segment audio files into
    vocalizations, and then predict labels for those segments.
    The function also prepares datasets so neural networks can predict the
    segmentation and annotation of vocalizations in them.
    It can also split a dataset into training, validation, and test sets,
    e.g. for benchmarking different neural network architectures.

    If no durations for any of the training sets are specified, then the
    function assumes all the vocalizations constitute a single training
    dataset. If the duration of either the training or test set is provided,
    then the function attempts to split the dataset into training and test sets.
    """
    toml_path = Path(toml_path)
    cfg = config.parse.from_toml(toml_path, sections=['PREP', 'SPECTROGRAM', 'DATALOADER'])

    if cfg.prep is None:
        raise ValueError(
            f'prep called with a config.ini file that does not have a PREP section: {toml_path}'
        )

    # pre-conditions ---------------------------------------------------------------------------------------------------
    if cfg.prep.audio_format is None and cfg.prep.spect_format is None:
        raise ValueError("Must specify either audio_format or spect_format")

    if cfg.prep.audio_format and cfg.prep.spect_format:
        raise ValueError("Cannot specify both audio_format and spect_format, "
                         "unclear whether to create spectrograms from audio files or "
                         "use already-generated spectrograms from array files")

    if cfg.prep.labelset is not None:
        if type(cfg.prep.labelset) not in (set, list):
            raise TypeError(
                f"type of labelset must be set or list, but type was: {type(labelset)}"
            )

        if type(cfg.prep.labelset) == list:
            labelset_set = set(cfg.prep.labelset)
            if len(cfg.prep.labelset) != len(labelset_set):
                raise ValueError(
                    'labelset contains repeated elements, should be a set (i.e. all members unique.\n'
                    f'Labelset was: {cfg.prep.labelset}'
                )
            else:
                cfg.prep.labelset = labelset_set

    if cfg.prep.output_dir:
        if not os.path.isdir(cfg.prep.output_dir):
            raise NotADirectoryError(
                f'output_dir not found: {cfg.prep.output_dir}'
            )
    elif cfg.prep.output_dir is None:
        cfg.prep.output_dir = cfg.prep.data_dir

    # ---- logging -----------------------------------------------------------------------------------------------------
    logger = logging.getLogger(__name__)
    logger.setLevel('INFO')

    # ---- figure out what section we will be saving path in, and prefix of option name in that section ----------------
    # (e.g., if it's PREDICT section then the prefix will be 'predict' for 'predict_vds_path' option
    with toml_path.open('r') as fp:
        config_toml = toml.load(fp)
    if 'TRAIN' in config_toml:
        section = 'TRAIN'
    elif 'LEARNCURVE' in config_toml:
        section = 'LEARNCURVE'
    elif 'PREDICT' in config_toml:
        section = 'PREDICT'
    else:
        raise ValueError(
            'Did not find a section named TRAIN, LEARNCURVE, or PREDICT in config.ini file;'
            ' unable to determine which section to add paths to prepared datasets to'
        )

    # ---- figure out file name ----------------------------------------------------------------------------------------
    timenow = datetime.now().strftime('%y%m%d_%H%M%S')
    _, tail = os.path.split(cfg.prep.data_dir)
    csv_fname_stem = f'{tail}_prep_{timenow}'
    csv_path = os.path.join(cfg.prep.output_dir, f'{csv_fname_stem}.csv')

    # ---- figure out if we're going to split into train / val / test sets ---------------------------------------------
    if all([dur is None for dur in (cfg.prep.train_dur,
                                    cfg.prep.val_dur,
                                    cfg.prep.test_dur)]):
        # then we're not going to split
        do_split = False
    else:
        if cfg.prep.val_dur is not None and cfg.prep.train_dur is None and cfg.prep.test_dur is None:
            raise ValueError('cannot specify only val_dur, unclear how to split dataset into training and test sets')
        else:
            do_split = True

    # ---- actually make the dataset -----------------------------------------------------------------------------------
    vak_df = dataframe.from_files(labelset=cfg.prep.labelset,
                                  data_dir=cfg.prep.data_dir,
                                  annot_format=cfg.prep.annot_format,
                                  output_dir=cfg.prep.output_dir,
                                  annot_file=cfg.prep.annot_file,
                                  audio_format=cfg.prep.audio_format,
                                  spect_format=cfg.prep.spect_format,
                                  spect_params=cfg.spect_params)

    if do_split:
        # save before splitting, jic duration args are not valid (we can't know until we make dataset)
        vak_df.to_csv(csv_path)
        vak_df = train_test_dur_split(vak_df,
                                      labelset=cfg.prep.labelset,
                                      train_dur=cfg.prep.train_dur,
                                      val_dur=cfg.prep.val_dur,
                                      test_dur=cfg.prep.test_dur)

    elif do_split is False:
        # add a split column, but assign everything to the same 'split'
        vak_df = dataframe.add_split_col(vak_df, split=section.lower())

    vak_df.to_csv(csv_path)

    # use config and section from above to add csv_path to config.ini file
    config_toml[section]['csv_path'] = csv_path

    with toml_path.open('w') as fp:
        toml.dump(config_toml, fp)



import crowsetta
import pandas as pd

from . import labels
from .path import array_dict_from_path


def has_unlabeled(csv_path, labelset, timebins_key='t'):
    """determine if a dataset has segments that are unlabeled.

    Used to decide whether an additional class needs to be added
    to the set of labels Y = {y_1, y_2, ... y_n}, where the added
    class y_n+1 will represent the unlabeled segments.

    Parameters
    ----------
    csv_path : str, Path
        to .csv file representing dataset.
    labelset : set
        of labels, str or int, e.g. {'a', 'b', 'c'}
    timebins_key : str
        key used to access timebins vector in spectrogram files.
        Default is 't'.

    Returns
    -------
    has_unlabeled : bool
        if True, dataset has unlabeled segments.
    """
    tmp_labelmap = labels.to_map(labelset, map_unlabeled=False)

    vak_df = pd.read_csv(csv_path)
    uniq_annot_formats = vak_df['annot_format'].unique()
    if len(uniq_annot_formats) == 1:
        has_only_one_annot_format = True
        annot_format = uniq_annot_formats.item()
        scribe = crowsetta.Transcriber(annot_format=annot_format)
    else:
        has_only_one_annot_format = False

    has_unlabeled_list = []
    for _, voc in vak_df.iterrows():
        if not has_only_one_annot_format:
            scribe = crowsetta.Transcriber(annot_format=voc['annot_format'])
        annot = scribe.from_file(voc['annot_path'])
        time_bins = array_dict_from_path(voc['spect_path'])[timebins_key]

        lbls_int = [tmp_labelmap[lbl] for lbl in annot.seq.labels]
        has_unlabeled_list.append(
            labels.has_unlabeled(lbls_int,
                                 annot.seq.onsets_s,
                                 annot.seq.offsets_s,
                                 time_bins)
        )

    return any(has_unlabeled_list)



"""parses [TRAIN] section of config"""
import attr
from attr import converters, validators
from attr.validators import instance_of

from .converters import bool_from_str, comma_separated_list, expanded_user_path
from .validators import is_a_directory, is_a_file, is_valid_model_name
from ..util.general import get_default_device


@attr.s
class TrainConfig:
    """class that represents [TRAIN] section of config.ini file

    Attributes
    ----------
    models : list
        comma-separated list of model names.
        e.g., 'models = TweetyNet, GRUNet, ConvNet'
    csv_path : str
        path to where dataset was saved as a csv.
    num_epochs : int
        number of training epochs. One epoch = one iteration through the entire
        training set.
    batch_size : int
        number of samples per batch presented to models during training.
    root_results_dir : str
        directory in which results will be created.
        The vak.cli.train function will create
        a subdirectory in this directory each time it runs.
    shuffle: bool
        if True, shuffle training data before each epoch. Default is True.
    normalize_spectrograms : bool
        if True, use spect.utils.data.SpectScaler to normalize the spectrograms.
        Normalization is done by subtracting off the mean for each frequency bin
        of the training set and then dividing by the std for that frequency bin.
        This same normalization is then applied to validation + test data.
    val_error_step : int
        step/epoch at which to estimate accuracy using validation set.
        Default is None, in which case no validation is done.
    checkpoint_step : int
        step/epoch at which to save to checkpoint file.
        Default is None, in which case checkpoint is only saved at the last epoch.
    patience : int
        number of epochs to wait without the error dropping before stopping the
        training. Default is None, in which case training continues for num_epochs
    save_only_single_checkpoint_file : bool
        if True, save only one checkpoint file instead of separate files every time
        we save. Default is True.
    """
    # required
    models = attr.ib(converter=comma_separated_list,
                     validator=[instance_of(list), is_valid_model_name])
    num_epochs = attr.ib(converter=int, validator=instance_of(int))
    batch_size = attr.ib(converter=int, validator=instance_of(int))
    root_results_dir = attr.ib(converter=expanded_user_path, validator=is_a_directory)

    # optional
    # csv_path is actually 'required' but we can't enforce that here because cli.prep looks at
    # what sections are defined to figure out where to add csv_path after it creates the csv
    csv_path = attr.ib(converter=converters.optional(expanded_user_path),
                       validator=validators.optional(is_a_file),
                       default=None
                       )

    results_dirname = attr.ib(converter=converters.optional(expanded_user_path),
                              validator=validators.optional(is_a_directory), default=None)

    normalize_spectrograms = attr.ib(converter=bool_from_str,
                                     validator=validators.optional(instance_of(bool)), default=False)

    num_workers = attr.ib(validator=instance_of(int), default=2)
    device = attr.ib(validator=instance_of(str), default=get_default_device())
    shuffle = attr.ib(converter=bool_from_str, validator=instance_of(bool), default=True)


    val_error_step = attr.ib(converter=converters.optional(int),
                             validator=validators.optional(instance_of(int)), default=None)
    checkpoint_step = attr.ib(converter=converters.optional(int),
                              validator=validators.optional(instance_of(int)), default=None)
    patience = attr.ib(converter=converters.optional(int),
                       validator=validators.optional(instance_of(int)), default=None)
    save_only_single_checkpoint_file = attr.ib(converter=bool_from_str,
                                               validator=instance_of(bool), default=True)


REQUIRED_TRAIN_OPTIONS = [
    'models',
    'root_results_dir',
]


def parse_train_config(config_toml, toml_path):
    """parse [TRAIN] section of config.toml file

    Parameters
    ----------
    config_toml : dict
        containing configuration file in TOML format, already loaded by parse function
    toml_path : Path
        path to a configuration file in TOML format (used for error messages)

    Returns
    -------
    train_config : vak.config.train.TrainConfig
        instance of TrainConfig class
    """
    train_section = config_toml['TRAIN']
    train_section = dict(train_section.items())
    for required_option in REQUIRED_TRAIN_OPTIONS:
        if required_option not in train_section:
            raise KeyError(
                f"the '{required_option}' option is required but was not found in the "
                f"TRAIN section of the config.ini file: {toml_path}"
            )
    return TrainConfig(**train_section)



from crowsetta import Transcriber
import numpy as np
import pandas as pd
import torch
from torchvision.datasets.vision import VisionDataset

from .. import util


class SpectrogramWindowDataset(VisionDataset):
    """Dataset class that represents a set of spectrograms of vocalizations
    and annotations for those vocalizations.

    Returns windows from the spectrograms, along with labels for each
    time bin in the window, derived from the annotations."""
    def __init__(self,
                 root,
                 x_inds,
                 spect_id_vector,
                 spect_inds_vector,
                 spect_paths,
                 annot_paths,
                 annot_formats,
                 scribes,
                 labelmap,
                 window_size,
                 spect_key='s',
                 timebins_key='t',
                 transform=None,
                 target_transform=None,
                 ):
        """initialize a SpectrogramWindowDataset instance

        Parameters
        ----------
        root : str, Path
            for SpectrogramWindowDataset, this should be a path
            to a .csv file that represents the dataset.
            Name 'root' is used for consistency with torchvision.datasets
        x_inds : numpy.ndarray
            indices of each window in the dataset
        spect_id_vector : numpy.ndarray
            represents the 'id' of any spectrogram,
            i.e., the index into spect_paths that will let us load it
        spect_inds_vector : numpy.ndarray
            valid indices of windows we can grab from each spectrogram
        spect_paths : numpy.ndarray
            column from DataFrame that represents dataset,
            consisting of paths to files containing spectrograms as arrays
        annot_paths : numpy.ndarray
            column from DataFrame that represents dataset,
            consisting of paths to annotation files
        annot_formats : numpy.ndarray
            column from DataFrame that represents dataset,
            consisting of annotation formats
        scribes : dict
            where keys are unique annotation formats from annot_formats,
            and values are crowsetta.Transcriber instances for those formats
        labelmap : dict
            that maps labels from dataset to a series of consecutive integer.
            To create a label map, pass a set of labels to the `vak.utils.labels.to_map` function.
        window_size : int
            number of time bins in windows that will be taken from spectrograms
        spect_key : str
            key to access spectograms in array files. Default is 's'.
        timebins_key : str
            key to access time bin vector in array files. Default is 't'.
        transform : callable
            A function/transform that takes in a numpy array
            and returns a transformed version. E.g, a SpectScaler instance.
            Default is None.
        target_transform : callable
            A function/transform that takes in the target and transforms it.
        """
        super(SpectrogramWindowDataset, self).__init__(root, transform=transform,
                                                       target_transform=target_transform)
        self.x_inds = x_inds
        self.spect_id_vector = spect_id_vector
        self.spect_inds_vector = spect_inds_vector
        self.spect_paths = spect_paths
        self.spect_key = spect_key
        self.timebins_key = timebins_key
        self.annot_paths = annot_paths
        self.annot_formats = annot_formats
        self.scribes = scribes
        self.labelmap = labelmap
        if 'unlabeled' in self.labelmap:
            self.unlabeled_label = self.labelmap['unlabeled']
        else:
            # if there is no "unlabeled label" (e.g., because all segments have labels)
            # just assign dummy value that will end up getting replaced by actual labels by label_timebins()
            self.unlabeled_label = 0
        self.window_size = window_size

        tmp_x_ind = 0
        one_x, _ = self.__getitem__(tmp_x_ind)
        # used by vak functions that need to determine size of window,
        # e.g. when initializing a neural network model
        self.shape = one_x.shape

    def __get_window_labelvec(self, idx):
        """helper function that gets batches of training pairs,
        given indices into dataset

        Parameters
        ----------
        idx : integer
            index into dataset

        Returns
        -------
        window : numpy.ndarray
            window from spectrograms
        labelvec : numpy.ndarray
            vector of labels for each timebin in window from spectrogram
        """
        x_ind = self.x_inds[idx]
        spect_id = self.spect_id_vector[x_ind]
        window_start_ind = self.spect_inds_vector[x_ind]

        spect_path = self.spect_paths[spect_id]
        spect_dict = util.path.array_dict_from_path(spect_path)
        spect = spect_dict[self.spect_key]
        timebins = spect_dict[self.timebins_key]

        annot_path = self.annot_paths[spect_id]
        annot_format = self.annot_formats[spect_id]
        annot = self.scribes[annot_format].from_file(annot_path)

        lbls_int = [self.labelmap[lbl] for lbl in annot.seq.labels]
        lbl_tb = util.labels.label_timebins(lbls_int,
                                            annot.seq.onsets_s,
                                            annot.seq.offsets_s,
                                            timebins,
                                            unlabeled_label=self.unlabeled_label)

        window = spect[:, window_start_ind:window_start_ind + self.window_size]
        labelvec = lbl_tb[window_start_ind:window_start_ind + self.window_size]

        return window, labelvec

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        window, labelvec = self.__get_window_labelvec(idx)

        if self.transform is not None:
            window = self.transform(window)

        if self.target_transform is not None:
            labelvec = self.target_transform(labelvec)

        return window, labelvec

    def __len__(self):
        """number of batches"""
        return len(self.x_inds)

    @classmethod
    def from_csv(cls, csv_path, split, labelmap, window_size,
                 spect_key='s', timebins_key='t', transform=None, target_transform=None):
        """given a path to a csv representing a dataset,
        returns an initialized SpectrogramWindowDataset.

        Parameters
        ----------
        csv_path : str, Path
            path to csv that represents dataset.
        split : str
            name of split from dataset to use
        labelmap : dict
            that maps labels from dataset to a series of consecutive integer.
            To create a label map, pass a set of labels to the `vak.utils.labels.to_map` function.
        window_size : int
            number of time bins in windows that will be taken from spectrograms
        spect_key : str
            key to access spectograms in array files. Default is 's'.
        timebins_key : str
            key to access time bin vector in array files. Default is 't'.
        transform : callable
            A function/transform that takes in a numpy array
            and returns a transformed version. E.g, a SpectScaler instance.
            Default is None.
        target_transform : callable
            A function/transform that takes in the target and transforms it.

        Returns
        -------
        initialized instance of SpectrogramWindowDataset
        """
        df = pd.read_csv(csv_path)
        if not df['split'].str.contains(split).any():
            raise ValueError(
                f'split {split} not found in dataset in csv: {csv_path}'
            )
        else:
            df = df[df['split'] == split]

        spect_paths = df['spect_path'].values

        def n_time_bins_spect(spect_path, spect_key=spect_key):
            spect = util.path.array_dict_from_path(spect_path)[spect_key]
            return spect.shape[-1]

        # to represent a dataset of windows from spectrograms without actually loading
        # all the spectrograms and concatenating them into one big matrix,
        # we will make three vectors that correspond to this imaginary, unloaded big matrix:
        # (1) `spect_id_vector` that represents the 'id' of any spectrogram in this matrix,
        # i.e., the index into spect_paths that will let us load it, and
        # (2) `spect_inds_vector` where the elements represents valid indices of windows
        # we can grab from each spectrogram. Valid indices are any up to the index n, where
        # n = number of time bins in this spectrogram - number of time bins in our window
        # (because if we tried to go past that the window would go past the edge of the
        # spectrogram). Lastly we make our 'training set' vector x, which is just a set
        # of indices (0, 1, ..., m) where m is the length of vectors (1) and (2).
        # When we want to grab a batch of size b of windows, we get b indices from x,
        # and then index into vectors (1) and (2) so we know which spectrogram files to
        # load, and which windows to grab from each spectrogram
        spect_id_vector = []  # tells us the index of spect_path
        spect_inds_vector = []  # tells us the index of valid windows in spect loaded from spect_path
        for ind, spect_path in enumerate(spect_paths):
            n_tb_spect = n_time_bins_spect(spect_path)
            # calculate number of windows we can extract from spectrogram of width time_bins
            n_windows = n_tb_spect - window_size
            spect_id_vector.append(np.ones((n_windows,), dtype=np.int64) * ind)
            spect_inds_vector.append(np.arange(n_windows))
        spect_id_vector = np.concatenate(spect_id_vector)
        spect_inds_vector = np.concatenate(spect_inds_vector)

        x_inds = np.arange(spect_id_vector.shape[0])

        annot_paths = df['annot_path'].values
        annot_formats = df['annot_format'].values
        uniq_formats = np.unique(annot_formats)
        scribes = {}
        for annot_format in uniq_formats:
            scribes[annot_format] = Transcriber(annot_format=annot_format)

        # note that we set "root" to csv path
        return cls(csv_path,
                   x_inds,
                   spect_id_vector,
                   spect_inds_vector,
                   spect_paths,
                   annot_paths,
                   annot_formats,
                   scribes,
                   labelmap,
                   window_size,
                   spect_key,
                   timebins_key,
                   transform,
                   target_transform
                   )



from collections import defaultdict

import torch
import torch.nn.modules.loss
import torch.optim
from tqdm import tqdm

from ..util.general import get_default_device


class Model:
    """lightweight model class that adds methods for training and evaluation
    to PyTorch neural networks.

    Sets forth a consistent API that models should implement to work with vak.

    Class Attributes
    ----------------
    network : torch.nn.Module
        subclass that implements a neural network
    loss : callable
        that implements a loss function, often called "criterion" in PyTorch code.
    optimizer : callable
        that optimizes model by iteratively updating trainable parameters.
    metrics : dict
        where keys are metric names, and values are callables that compute that a metric,
        e.g. accuracy. Metrics should accept arguments y_pred and y_true.

    Attributes
    ----------
    device : str
        device on which to place tensors. One of {"cuda", "cpu}.

    Methods
    -------
    fit : fit a model by training it with supplied data for a specified number of epochs
    evaluate : evaluate a model by computing specified metrics on supplied data
    predict : return predictions of model, i.e. output when fed with supplied data
    compile : returns instance of model with attributes set to specified arguments

    Private Methods
    ---------------
    _train : helper method, called by the fit method on each epoch.
        Fits the model by iterating once through all the training data.
        Override this method if you need to implement your own training method.
    _eval : helper method, called by the evaluate method, and called by the fit
        method for validation after each epoch. Evaluates the model by iterating
        through eval_data and computing the specified metrics for each batch.
        Override this method if you need to implement your own validation and test method.
    _predict : helper method, called by the predict method on each epoch.
        Uses the model to make predictions, by iterating through pred_data
        and returning the outputs of each batch fed into it.
        Override this method if you need to implement your own predict method.
    """

    REQUIRED_SUBCLASS_ATTRIBUTES = [
        'network',
        'optimizer',
        'loss',
        'metrics',
    ]

    def __init__(self,
                 network,
                 loss,
                 optimizer,
                 metrics):
        self.network = network
        self.optimizer = optimizer
        self.loss = loss
        self.metrics = metrics
        self.device = None

    def _train(self, train_data):
        """helper method, called by the fit method on each epoch.
        Iterates once through train_data, using it to update model parameters.
        Override this method if you need to implement your own training method.

        Parameters
        ----------
        train_data : torch.util.Dataloader
            instance that will be iterated over.
        """
        self.network.train()

        progress_bar = tqdm(train_data)
        for ind, batch in enumerate(progress_bar):
            x, y = batch[0].to(self.device), batch[1].to(self.device)
            y_pred = self.network.forward(x)
            self.optimizer.zero_grad()
            loss = self.loss(y_pred, y)
            loss.backward()
            self.optimizer.step()
            progress_bar.set_description(
                f'batch {ind}, loss: {loss.item():.4f}'
            )

    def _eval(self, eval_data):
        """helper method, called by the evaluate method, and called by the fit
        method for validation after each epoch. Evaluates the model by iterating
        through eval_data and computing the specified metrics for each batch.
        Override this method if you need to implement your own validation and test method.

        Parameters
        ----------
        eval_data : torch.util.Dataloader
            instance that will be iterated over.
        """
        self.network.eval()

        metric_vals = defaultdict(list)

        n_batches = 0

        progress_bar = tqdm(eval_data)
        with torch.no_grad():
            for ind, batch in enumerate(progress_bar):
                x, y = batch[0].to(self.device), batch[1].to(self.device)
                y_pred = self.network.forward(x)
                for metric_name, metric_callable in self.metrics.items():
                    metric_vals[metric_name].append(
                        metric_callable(y_pred, y)
                    )
                n_batches += 1
                progress_bar.set_description(
                    f'batch {ind} / {len(eval_data)}'
                )

        for metric_name in metric_vals.keys():
            if metric_name in ['loss', 'acc']:
                avg_metric_val = (
                        torch.tensor(metric_vals[metric_name]).sum() / n_batches
                ).item()
                metric_vals[metric_name] = avg_metric_val
            else:
                raise NotImplementedError(
                    f'calculation of metric across batches not yet implemented for {metric_name}'
                )

        return metric_vals

    def _predict(self, pred_data):
        """helper method, called by the predict method on each epoch.
        Uses the model to make predictions, by iterating through pred_data
        and returning the outputs of each batch fed into it.
        Override this method if you need to implement your own predict method.

        Parameters
        ----------
        pred_data : torch.util.Dataloader
            instance that will be iterated over.
        """
        self.network.eval()

        y_pred_all = []

        progress_bar = tqdm(pred_data)

        with torch.no_grad():
            for ind, batch in enumerate(progress_bar):
                x, y = batch[0].to(self.device), batch[1].to(self.device)
                y_pred = self.network.forward(x)
                y_pred_all.append(y_pred)

                progress_bar.set_description(
                    f'batch {ind} / {len(pred_data)}'
                )

        return torch.cat(y_pred_all)

    def save(self, ckpt_path, epoch, **kwargs):
        """save model state to a checkpoint file.

        Saves epoch, network state_dict, optimizer state_dict, and
        any keyword arguments to a checkpoint file with the name
        specified by ckpt_path.

        Parameters
        ----------
        ckpt_path : str, Path
            path including filename that should be used to save checkpoint
        epoch : int
            epoch on which this checkpoint was saved
        """
        ckpt = {
            'epoch': epoch,
            'network_state_dict': self.network.state_dict(),
            'optimizer_state_dict': self.optimizer.state_dict(),
        }
        ckpt.update(**kwargs)
        torch.save(ckpt, ckpt_path)

    def load(self, ckpt_path):
        """load model state dict from a checkpoint file.

        Loads state_dicts into network and optimizer
        from a checkpoint file with the name
        specified by ckpt_path.

        Parameters
        ----------
        ckpt_path : str, Path
            path including filename from which to load checkpoint
        """
        ckpt = torch.load(ckpt_path)
        self.network.load_state_dict(ckpt['network_state_dict'])
        self.optimizer.load_state_dict(ckpt['optimizer_state_dict'])

    def fit(self,
            train_data,
            num_epochs,
            ckpt_root,
            val_data=None,
            val_step=1,
            checkpoint_step=1,
            save_best_only=True,
            patience=None,
            single_ckpt=True,
            device=None
            ):
        if save_best_only:
            if val_data is None:
                raise ValueError(
                    'save_best_only is True but no validation dataset was provided to measure accuracy'
                )

        if patience is not None:
            if val_data is None:
                raise ValueError(
                    f'patience set to {patience}, but no validation dataset was provided to measure accuracy'
                )

        if device is None:
            device = get_default_device()
        self.device = device

        if save_best_only or patience:
            max_val_acc = 0

        if patience:
            patience_counter = 0

        self.network.to(self.device)

        for epoch in range(1, num_epochs + 1):
            print(f'epoch {epoch} / {num_epochs}')
            self._train(train_data)

            if single_ckpt:
                ckpt_path = ckpt_root.joinpath('checkpoint.pt')
            else:
                ckpt_path = ckpt_root.joinpath(f'checkpoint_epoch{epoch}.pt')

            if val_data is not None:
                if epoch % val_step == 0:
                    print('computing metrics on validation set')
                    metric_vals = self._eval(val_data)
                    print(
                        ', '.join([f'{k}: {v:.4f}' for k, v in metric_vals.items()])
                    )

                    if patience or epoch % checkpoint_step == 0:
                        epoch_acc = metric_vals['acc']
                        if patience:
                            if epoch_acc > max_val_acc:
                                max_val_acc = metric_vals['acc']
                                patience_counter = 0
                                print(
                                    f'accuracy improved, saving checkpoint'
                                )
                                self.save(ckpt_path, epoch)
                            else:
                                patience_counter += 1
                                if patience_counter > patience:
                                    print(
                                        f'early stopping, validation accuracy has not improved in {patience} epochs')
                                    if not save_best_only:
                                        self.save(ckpt_path, epoch)
                                break
                        elif epoch % checkpoint_step == 0:
                            if save_best_only:
                                if epoch_acc < max_val_acc:
                                    continue
                            self.save(ckpt_path, epoch)

            elif epoch % checkpoint_step == 0:  # but we don't have validation data
                self.save(ckpt_path, epoch)

    def evaluate(self,
                 eval_data,
                 device=None):
        if device is None:
            device = get_default_device()
        self.device = device
        self.network.to(self.device)
        return self._eval(eval_data)

    def predict(self,
                pred_data,
                device=None):
        if device is None:
            device = get_default_device()
        self.device = device
        self.network.to(self.device)
        return self._predict(pred_data)

    @classmethod
    def from_config(cls, config):
        """any model that inherits from this class should do whatever it needs to
        in this factory method to create the network, optimizer, and loss, and
        then pass those to the init function

        Parameters
        ----------
        config : dict
            presumably mapping 'network', 'optimizer' and 'loss' to kwargs that
            determine parameters for each of those, e.g. dropout rate for the
            network, learning rate for the optimizer, etc.

        Returns
        -------
        model : Model
            instance of a model that subclasses this Model
        """
        raise NotImplementedError



from . import audio
from . import curvefit
from . import data
from . import dataset
from . import entry_points
from . import general
from . import labels
from . import path
from . import splitalgos
from . import spect
from . import validation

from .split import train_test_dur_split



"""utility functions specific to engine sub-package"""
BUILTIN_TYPES = [dict, list, str, set, tuple]

# define this outside Model class, so sub-classes can use it too
def _check_required_attribs(self, cls):
    """helper function that checks whether subclass declares attributes required by supraclass

    Parameters
    ----------
    self : Model
        or Model subclass that calls this function
        to check that a subclass of it has required attributes
    cls : type
        subclass of self, that must have the required class attributes
    """
    if not hasattr(self, 'REQUIRED_SUBCLASS_ATTRIBUTES'):
        raise AttributeError(
            f'Unable to validate whether {cls.__name__} has required attributes, '
            f'because {self.__name__} does not have the class variable "REQUIRED_SUBCLASS_ATTRIBUTES"'
        )
    for required_subclass_attr, subclass_attr_type in self.REQUIRED_SUBCLASS_ATTRIBUTES.items():
        if not hasattr(cls, required_subclass_attr):
            raise AttributeError(
                f'A subclass of {self.__name__} must declare a {required_subclass_attr},'
                f' but {cls.__name__} does not have an attribute {required_subclass_attr}.'
            )

        attr = getattr(cls, required_subclass_attr)
        if subclass_attr_type in BUILTIN_TYPES:
            # if it's required attr has a built-in data type, not a custom Python code class,
            # check if the attr is an instance of that built-in data type
            if type(attr) != subclass_attr_type:
                raise TypeError(
                    f'type of {attr} should be {subclass_attr_type} but was {type(attr)}'
                )
        else:
            if not issubclass(attr, subclass_attr_type):
                raise TypeError(
                    f'{required_subclass_attr} must be a subclass of {subclass_attr_type}, but '
                    f'type was: {type(attr)}'
                )



from pathlib import Path
from distutils.util import strtobool


def bool_from_str(value):
    if type(value) == bool:
        return value
    elif type(value) == str:
        return bool(strtobool(value))


def comma_separated_list(value):
    if type(value) is list:
        return value
    elif type(value) is str:
        return [element.strip() for element in value.split()]
    else:
        raise TypeError(
            f'unexpected type when converting to comma-separated list: {type(value)}'
        )


def expanded_user_path(value):
    return Path(value).expanduser()




import copy
import itertools
import os
import random
from glob import glob

import joblib
import numpy as np
from tqdm import tqdm


def make_data_dicts(output_dir,
                    total_train_set_duration,
                    validation_set_duration,
                    test_set_duration,
                    labelset,
                    spect_files=None):
    """function that loads data and saves in dictionaries

    Parameters
    ----------
    output_dir : str
        path to output_dir containing .spect files
    total_train_set_duration : float
    validation_set_duration : float
    test_set_duration : float
        all in seconds
    labelset : list
        of str, labels used
    spect_files : str
        full path to file containing 'spect_files' list of tuples
        saved by function make_spects_from_list_of_files.
        Default is None, in which case this function looks for
        a file named 'spect_files' in output_dir.

    Returns
    -------
    saved_data_dict_paths : dict
        with keys {'train', 'test', 'val'} and values being the path
        to which the data_dict was saved

    Saves three 'data_dict' files (train, validation, and test)
    in output_dir, with following structure:
        spects : list
            of ndarray, spectrograms from audio files
        filenames : list
            same length as spects, filename of each audio file that was converted to spectrogram
        freq_bins : ndarray
            vector of frequencies where each value is a bin center. Same for all spectrograms
        time_bins : list
            of ndarrays, each a vector of times where each value is a bin center.
            One for each spectrogram
        labelset : list
            of strings, labels corresponding to each spectrogram
        labeled_timebins : list
            of ndarrays, each same length as time_bins but value is a label for that bin.
            In other words, the labels vector is mapped onto the time_bins vector for the
            spectrogram.
        X : ndarray
            X_train, X_val, or X_test, depending on which data_dict you are looking at.
            Some number of spectrograms concatenated, enough so that the total duration
            of the spectrogram in time bins is equal to or greater than the target duration.
            If greater than target, then X is truncated so it is equal to the target.
        Y : ndarray
            Concatenated labeled_timebins vectors corresponding to the spectrograms in X.
        spect_ID_vector : ndarray
            Vector where each element is an ID for a song. Used to randomly grab subsets
            of data of a target duration while still having the subset be composed of
            individual songs as much as possible. So this vector will look like:
            [0, 0, 0, ..., 1, 1, 1, ... , n, n, n] where n is equal to or (a little) less
            than the length of spects. spect_ID_vector.shape[-1] is the same as X.shape[-1]
            and Y.shape[0].
        timebin_dur : float
            duration of a timebin in seconds from spectrograms
        spect_params : dict
            parameters for computing spectrogram as specified in config.ini file.
            Will be checked against .ini file when running other cli such as learn_curve.py
        labels_mapping : dict
            maps str labels for syllables to consecutive integers.
            As explained in docstring for make_spects_from_list_of_files.
    """
    if not os.path.isdir(output_dir):
        raise NotADirectoryError('{} not recognized '
                                 'as a directory'.format(output_dir))

    if spect_files is None:
        spect_files = glob(os.path.join(output_dir,'spect_files'))
        if spect_files == []:  # if glob didn't find anything
            raise FileNotFoundError("did not find 'spect_files' file in {}"
                                    .format(output_dir))
        elif len(spect_files) > 1:
            raise ValueError("found than more than one 'spect_files' in {}:\n{}"
                             .format(output_dir, spect_files))
        else:
            spect_files = spect_files[0]

    if not os.path.isfile(spect_files):
        raise FileNotFoundError('{} not recognized as a file'
                                .format(spect_files))

    spect_files = joblib.load(spect_files)

    total_spects_dur = sum([spect[1] for spect in spect_files])
    total_dataset_dur = sum([total_train_set_duration,
                             validation_set_duration,
                             test_set_duration])
    if total_spects_dur < total_dataset_dur:
        raise ValueError('Total duration of all .cbin files, {} seconds,'
                         ' is less than total target duration of '
                         'training, validation, and test sets, '
                         '{} seconds'
                         .format(total_spects_dur, total_dataset_dur))

    # main loop that gets datasets
    iter = 1
    all_labels_err = ('Did not successfully divide data into training, '
                      'validation, and test sets of sufficient duration '
                      'after 1000 iterations.'
                      ' Try increasing the total size of the data set.')

    while 1:
        spect_files_copy = copy.deepcopy(spect_files)

        train_spects = []
        val_spects = []
        test_spects = []

        total_train_dur = 0
        val_dur = 0
        test_dur = 0

        choice = ['train', 'val', 'test']

        while 1:
            # pop tuples off cbins_used list and append to randomly-chosen
            # list, either train, val, or test set.
            # Do this until the total duration for each data set is equal
            # to or greater than the target duration for each set.
            try:
                ind = random.randint(0, len(spect_files_copy)-1)
            except ValueError:
                if len(spect_files_copy) == 0:
                    print('Ran out of spectrograms while dividing data into training, '
                          'validation, and test sets of specified durations. Iteration {}'
                          .format(iter))
                    iter += 1
                    break  # do next iteration
                else:
                    raise
            a_spect = spect_files_copy.pop(ind)
            which_set = random.randint(0, len(choice)-1)
            which_set = choice[which_set]
            if which_set == 'train':
                train_spects.append(a_spect)
                total_train_dur += a_spect[1]  # ind 1 is duration
                if total_train_dur >= total_train_set_duration:
                    choice.pop(choice.index('train'))
            elif which_set == 'val':
                val_spects.append(a_spect)
                val_dur += a_spect[1]  # ind 1 is duration
                if val_dur >= validation_set_duration:
                    choice.pop(choice.index('val'))
            elif which_set == 'test':
                test_spects.append(a_spect)
                test_dur += a_spect[1]  # ind 1 is duration
                if test_dur >= test_set_duration:
                    choice.pop(choice.index('test'))

            if len(choice) < 1:
                if np.sum(total_train_dur +
                                  val_dur +
                                  test_dur) < total_dataset_dur:
                    raise ValueError('Loop to find subsets completed but '
                                     'total duration of subsets is less than '
                                     'total duration specified by config file.')
                else:
                    break

            if iter > 1000:
                raise ValueError('Could not find subsets of sufficient duration in '
                                 'less than 1000 iterations.')

        # make sure no contamination between data sets.
        # If this is true, each set of filenames should be disjoint from others
        train_spect_files = [tup[0] for tup in train_spects]  # tup = a tuple
        val_spect_files = [tup[0] for tup in val_spects]
        test_spect_files = [tup[0] for tup in test_spects]
        assert set(train_spect_files).isdisjoint(val_spect_files)
        assert set(train_spect_files).isdisjoint(test_spect_files)
        assert set(val_spect_files).isdisjoint(test_spect_files)

        # make sure that each set contains all classes we
        # want the network to learn
        train_labels = itertools.chain.from_iterable(
            [spect[2] for spect in train_spects])
        train_labels = set(train_labels)  # make set to get unique values

        val_labels = itertools.chain.from_iterable(
            [spect[2] for spect in val_spects])
        val_labels = set(val_labels)

        test_labels = itertools.chain.from_iterable(
            [spect[2] for spect in test_spects])
        test_labels = set(test_labels)

        if train_labels != set(labelset):
            iter += 1
            if iter > 1000:
                raise ValueError(all_labels_err)
            else:
                print('Train labels did not contain all labels in labelset. '
                      'Getting new training set. Iteration {}'
                      .format(iter))
                continue
        elif val_labels != set(labelset):
            iter += 1
            if iter > 1000:
                raise ValueError(all_labels_err)
            else:
                print('Validation labels did not contain all labels in labelset. '
                      'Getting new validation set. Iteration {}'
                      .format(iter))
                continue
        elif test_labels != set(labelset):
            iter += 1
            if iter > 1000:
                raise ValueError(all_labels_err)
            else:
                print('Test labels did not contain all labels in labelset. '
                      'Getting new test set. Iteration {}'
                      .format(iter))
                continue
        else:
            break

    saved_data_dict_paths = {}

    for dict_name, spect_list, target_dur in zip(['train','val','test'],
                                                 [train_spects,val_spects,test_spects],
                                                 [total_train_set_duration,
                                                  validation_set_duration,
                                                  test_set_duration]):

        spects = []
        filenames = []
        all_time_bins = []
        labels = []
        labeled_timebins = []
        spect_ID_vector = []

        for spect_ind, spect_file in enumerate(spect_list):
            spect_dict = joblib.load(spect_file[0])
            spects.append(spect_dict['spect'])
            filenames.append(spect_file[0])
            all_time_bins.append(spect_dict['time_bins'])
            labels.append(spect_dict['labels'])
            labeled_timebins.append(spect_dict['labeled_timebins'])
            spect_ID_vector.extend([spect_ind] * spect_dict['time_bins'].shape[-1])

            if 'freq_bins' in locals():
                assert np.array_equal(spect_dict['freq_bins'], freq_bins)
            else:
                freq_bins = spect_dict['freq_bins']

            if 'labels_mapping' in locals():
                assert spect_dict['labels_mapping'] == labels_mapping
            else:
                labels_mapping = spect_dict['labels_mapping']

            if 'timebin_dur' in locals():
                assert spect_dict['timebin_dur'] == timebin_dur
            else:
                timebin_dur = spect_dict['timebin_dur']

            if 'spect_params' in locals():
                assert spect_dict['spect_params'] == spect_params
            else:
                spect_params = spect_dict['spect_params']

        X = np.concatenate(spects, axis=1)
        Y = np.concatenate(labeled_timebins)
        spect_ID_vector = np.asarray(spect_ID_vector, dtype='int')
        assert X.shape[-1] == Y.shape[0]  # Y has shape (timebins, 1)
        if X.shape[-1] > target_dur / timebin_dur:
            correct_length = np.round(target_dur / timebin_dur).astype(int)
            X = X[:, :correct_length]
            Y = Y[:correct_length, :]
            spect_ID_vector = spect_ID_vector[:correct_length]

        data_dict = {'spects': spects,
                     'filenames': filenames,
                     'freq_bins': freq_bins,
                     'time_bins': all_time_bins,
                     'labels': labels,
                     'labeled_timebins': labeled_timebins,
                     'spect_ID_vector': spect_ID_vector,
                     'X_' + dict_name: X,
                     'Y_' + dict_name: Y,
                     'timebin_dur': timebin_dur,
                     'spect_params': spect_params,
                     'labels_mapping': labels_mapping}

        print('saving data dictionary in {}'.format(output_dir))
        data_dict_path = os.path.join(output_dir, dict_name + '_data_dict')
        joblib.dump(data_dict, data_dict_path)
        saved_data_dict_paths[dict_name] = data_dict_path

    return saved_data_dict_paths


def make_data_dict_from_spect_files(labelset, spect_files='./spect_files',
                                    output_dir='.', dict_name='test'):
    """load data from a list of spect files and save into a dictionary.
    For running a summary of prediction error on a list of files you choose.

    Parameters
    ----------
    labelset : list
        of str, labels used
    spect_files : str
        full path to file containing 'spect_files' list of tuples
        saved by function make_spects_from_list_of_files.
        Default is None, in which case this function looks for
        a file named 'spect_files' in output_dir.
    output_dir : str
        path to output_dir containing .spect files. Default is '.' (current
        working directory).
    dict_name : str
        prefix to filename '_data_dict' in which the dictionary is saved.
        Default is 'test'.

    Returns
    -------
    data_dict_path

    Function saves 'data_dict' file in output_dir, with following structure:
        spects : list
            of ndarray, spectrograms from audio files
        filenames : list
            same length as spects, filename of each audio file that was converted to spectrogram
        freq_bins : ndarray
            vector of frequencies where each value is a bin center. Same for all spectrograms
        time_bins : list
            of ndarrays, each a vector of times where each value is a bin center.
            One for each spectrogram
        labelset : list
            of strings, labels corresponding to each spectrogram
        labeled_timebins : list
            of ndarrays, each same length as time_bins but value is a label for that bin.
            In other words, the labels vector is mapped onto the time_bins vector for the
            spectrogram.
        X : ndarray
            X_train, X_val, or X_test, depending on which data_dict you are looking at.
            Some number of spectrograms concatenated, enough so that the total duration
            of the spectrogram in time bins is equal to or greater than the target duration.
            If greater than target, then X is truncated so it is equal to the target.
        Y : ndarray
            Concatenated labeled_timebins vectors corresponding to the spectrograms in X.
        spect_ID_vector : ndarray
            Vector where each element is an ID for a song. Used to randomly grab subsets
            of data of a target duration while still having the subset be composed of
            individual songs as much as possible. So this vector will look like:
            [0, 0, 0, ..., 1, 1, 1, ... , n, n, n] where n is equal to or (a little) less
            than the length of spects. spect_ID_vector.shape[-1] is the same as X.shape[-1]
            and Y.shape[0].
        timebin_dur : float
            duration of a timebin in seconds from spectrograms
        spect_params : dict
            parameters for computing spectrogram as specified in config.ini file.
            Will be checked against .ini file when running other cli such as learn_curve.py
        labels_mapping : dict
            maps str labels for syllables to consecutive integers.
            As explained in docstring for make_spects_from_list_of_files.
    """
    spect_list = joblib.load(spect_files)
    spect_files = [tup[0] for tup in spect_list]

    labels = itertools.chain.from_iterable(
        [tup[2] for tup in spect_list])
    labels = set(labels)

    if labels != set(labelset):
        raise ValueError(
            'labels from all spect_files are not consistent with '
            'labels in labelset.')

    spects = []
    filenames = []
    all_time_bins = []
    labels = []
    labeled_timebins = []
    spect_ID_vector = []

    for spect_ind, tup in enumerate(spect_list):
        spect_dict = joblib.load(tup[0])
        spects.append(spect_dict['spect'])
        filenames.append(tup[0])
        all_time_bins.append(spect_dict['time_bins'])
        labels.append(spect_dict['labels'])
        labeled_timebins.append(spect_dict['labeled_timebins'])
        spect_ID_vector.extend([spect_ind] * spect_dict['time_bins'].shape[-1])

        if 'freq_bins' in locals():
            assert np.array_equal(spect_dict['freq_bins'], freq_bins)
        else:
            freq_bins = spect_dict['freq_bins']

        if 'labels_mapping' in locals():
            assert spect_dict['labels_mapping'] == labels_mapping
        else:
            labels_mapping = spect_dict['labels_mapping']

        if 'timebin_dur' in locals():
            assert spect_dict['timebin_dur'] == timebin_dur
        else:
            timebin_dur = spect_dict['timebin_dur']

        if 'spect_params' in locals():
            assert spect_dict['spect_params'] == spect_params
        else:
            spect_params = spect_dict['spect_params']

    X = np.concatenate(spects, axis=1)
    Y = np.concatenate(labeled_timebins)
    spect_ID_vector = np.asarray(spect_ID_vector, dtype='int')
    assert X.shape[-1] == Y.shape[0]  # Y has shape (timebins, 1)

    data_dict = {'spects': spects,
                 'filenames': filenames,
                 'freq_bins': freq_bins,
                 'time_bins': all_time_bins,
                 'labels': labels,
                 'labeled_timebins': labeled_timebins,
                 'spect_ID_vector': spect_ID_vector,
                 'X_' + dict_name: X,
                 'Y_' + dict_name: Y,
                 'timebin_dur': timebin_dur,
                 'spect_params': spect_params,
                 'labels_mapping': labels_mapping}

    data_dict_path = os.path.join(output_dir, dict_name + '_data_dict')
    print('saving data dictionary {} in {}'
          .format(data_dict_path, os.path.abspath(output_dir)))
    joblib.dump(data_dict, data_dict_path)
    return data_dict_path


def get_inds_for_dur(spect_ID_vector,
                     labeled_timebins_vector,
                     labels_mapping,
                     target_duration,
                     timebin_dur_in_s=0.001,
                     max_iter=1000,
                     method='incfreq'):
    """Randomly draw a subset of data with a specific duration.
    Draws spectrograms at random and adds to list until total duration of
    all spectrograms => target_duration, and then truncates at target duration.
    Returns vector of indices to select randomly-drawn subset from total dataset.

    Parameters
    ----------
    spect_ID_vector : ndarray
        vector where each element is an ID for a song. Used to randomly grab subsets
        of data of a target duration while still having the subset be composed of
        individual spectrograms as much as possible. So this vector will look like:
        [0, 0, 0, ..., 1, 1, 1, ... , n, n, n] where each n is the ID for one
        spectrogram, i.e. one audio file.
        spect_ID_vector.shape[-1] is the same as spectrograms.shape[-1] and labels.shape[0].
    labeled_timebins_vector : numpy.ndarray
        vector of same length as spect_ID_vector but each element is a label for the timebin,
        i.e. one of the possible classes.
    labels_mapping : dict
        maps string labels to consecutive integers. Used to check that the randomly drawn
        data set contains all classes.
    target_duration : float
        target duration of training set in seconds.
    timebin_dur_in_s : float
        duration of each timebin, i.e. each column in spectrogram, in seconds.
        Default is 0.001 s (1 ms)
    max_iter : int
        number of iterations to try drawing random subset of song that contains
        all classes in labels mapping.
        Default is 1000.
    method : str
        {'rand', 'incfreq'}
        method by which to obtain subset from training set.
        'incfreq' grabs songs at random but starts from the subset
        that includes the least frequently occurring class. Continues
        to grab randomly in order of increasing frequency until all
        classes are present, and then goes back to 'rand' method.
        'rand' grabs songs totally at random.
        Default is 'incfreq'.

    Returns
    -------
    inds_to_use : numpy.ndarray
        numpy array of indices, to index into rows of X_train that belong to chosen subset
        (assumes X_train is one long spectrogram, consisting of all
        training spectrograms concatenated, and each row being one timebin)
    """
    labeled_timebins_vector = np.squeeze(labeled_timebins_vector)

    if labeled_timebins_vector.ndim > 1:
        raise ValueError('number of dimensions of labeled_timebins_vector should be 1 '
                         '(after np.squeeze), but was equal to {}'
                         .format(labeled_timebins_vector.ndim))

    iter = 1
    with tqdm(total=max_iter) as pbar:
        while 1:  # keep iterating until we randomly draw subset that meets our criteria
            pbar.set_description(
                f'Randomly drawing subset of training data, attempt {iter} of {max_iter}'
            )
            if 'inds_to_use' in locals():
                del inds_to_use

            spect_IDs, spect_timebins = np.unique(spect_ID_vector, return_counts=True)

            if iter == 1:
                # sanity check:
                # spect_IDs should always start from 0
                # and go to n-1 where n is # of spectrograms
                assert np.array_equal(spect_IDs, np.arange(spect_IDs.shape[-1]))

            spect_IDs = spect_IDs.tolist()  # because we need to pop off ids for 'incfreq'

            spect_IDs_in_subset = []
            total_dur_in_timebins = 0

            if method == 'incfreq':
                classes, counts = np.unique(labeled_timebins_vector, return_counts=True)
                int_labels_without_int_flag = [val for val in labels_mapping.values()
                                               if type(val) is int]
                if set(classes) != set(int_labels_without_int_flag):
                    raise ValueError('classes in labeled_timebins_vector '
                                     'do not match classes in labels_mapping.')
                freq_rank = np.argsort(counts).tolist()

                # reason for doing it in this Schliemel-the-painter-looking way is that
                # I want to make sure all classes are represented first, but then
                # go back to just grabbing songs completely at random
                while freq_rank:  # is not an empty list yet
                    curr_class = classes[freq_rank.pop(0)]

                    # if curr_class already represented in subset, skip it
                    if 'inds_to_use' in locals():
                        classes_already_in_subset = np.unique(
                            labeled_timebins_vector[inds_to_use])
                        if curr_class in classes_already_in_subset:
                            continue

                    inds_this_class = np.where(labeled_timebins_vector==curr_class)[0]
                    spect_IDs_this_class = np.unique(spect_ID_vector[inds_this_class])
                    # keep only the spect IDs we haven't popped off main list already
                    spect_IDs_this_class = [spect_ID_this_class
                                            for spect_ID_this_class in spect_IDs_this_class
                                            if spect_ID_this_class in spect_IDs]
                    rand_spect_ID = np.random.choice(spect_IDs_this_class)

                    spect_IDs_in_subset.append(rand_spect_ID)
                    spect_IDs.pop(spect_IDs.index(rand_spect_ID))  # so as not to reuse it
                    # below, [0] because np.where returns tuple
                    spect_ID_inds = np.where(spect_ID_vector == rand_spect_ID)[0]
                    if 'inds_to_use' in locals():
                        inds_to_use = np.concatenate((inds_to_use, spect_ID_inds))
                    else:
                        inds_to_use = spect_ID_inds
                    total_dur_in_timebins += spect_timebins[rand_spect_ID]
                    if total_dur_in_timebins * timebin_dur_in_s >= target_duration:
                        # if total_dur greater than target, need to truncate
                        if total_dur_in_timebins * timebin_dur_in_s > target_duration:
                            correct_length = np.round(target_duration / timebin_dur_in_s).astype(int)
                            inds_to_use = inds_to_use[:correct_length]
                        # (if equal to target, don't need to do anything)
                        break

            if method=='rand' or \
                    total_dur_in_timebins * timebin_dur_in_s < target_duration:

                shuffled_spect_IDs = np.random.permutation(spect_IDs)

                for spect_ID in shuffled_spect_IDs:
                    spect_IDs_in_subset.append(spect_ID)
                    # below, [0] because np.where returns tuple
                    spect_ID_inds = np.where(spect_ID_vector == spect_ID)[0]
                    if 'inds_to_use' in locals():
                        inds_to_use = np.concatenate((inds_to_use, spect_ID_inds))
                    else:
                        inds_to_use = spect_ID_inds
                    total_dur_in_timebins += spect_timebins[spect_ID]
                    if total_dur_in_timebins * timebin_dur_in_s >= target_duration:
                        # if total_dur greater than target, need to truncate
                        if total_dur_in_timebins * timebin_dur_in_s > target_duration:
                            correct_length = np.round(target_duration / timebin_dur_in_s).astype(int)
                            inds_to_use = inds_to_use[:correct_length]
                        # (if equal to target, don't need to do anything)
                        break

            if set(np.unique(labeled_timebins_vector[inds_to_use])) != set(labels_mapping.values()):
                pbar.update(1)
                iter += 1

                if iter > max_iter:
                    raise ValueError('Attempted to draw subset of songs at random '
                                     'that contained all classes '
                                     'more than {} times unsuccessfully. Make sure '
                                     'that all classes are present in training data '
                                     'set.'.format(max_iter))
            else:
                return inds_to_use


def reshape_data_for_batching(X, batch_size, time_steps, Y=None):
    """Reshape data to feed to network in batches.
    Data is returned with shape (batch_size, num_batches * time_steps, -1)
    so that a batch can be grabbed at random starting at any index on axis 1
    (that is less than "length of axis 1 minus time steps").
    Pads data with zeros if it cannot be evenly divided into batches of size batch_size.

    Parameters
    ----------
    X : numpy.ndarray
        2-d matrix, concatenated spectrograms.
        Spectrograms are oriented so rows are time bins and columns are frequency bins.
    batch_size : int
        Number of samples in a batch
    time_steps : int
        Number of time steps in each sample, i.e., number of rows
    Y : numpy.ndarray
        vector of labels with shape (X.shape[0], 1). Default is None, for the case when
        you only have spectrograms in X and need to reshape them
        so you can predict labels Y.

    Returns
    -------
    X : numpy.ndarray
        Spectrograms reshaped to have shape (batch_size, num_batches * time_steps, -1).
    Y : numpy.ndarray
        Labels reshaped, will have shape (batch_size, -1). Only returned if Y is provided as an argument.
    num_batches : int
        num_batches = X.shape[0] // batch_size // time_steps (plus 1, if zero-padding was required).
    """
    num_batches = X.shape[0] // batch_size // time_steps
    rows_to_append = ((num_batches + 1) * time_steps * batch_size) - X.shape[0]

    if rows_to_append > 0:
        X = np.concatenate((X, np.zeros((rows_to_append, X.shape[1]))),
                           axis=0)
        if Y is not None:
            Y = np.concatenate((Y, np.zeros((rows_to_append, 1), dtype=int)), axis=0)
        num_batches = num_batches + 1

    X = X.reshape((batch_size, num_batches * time_steps, -1))
    if Y is not None:
        Y = Y.reshape((batch_size, -1))

    if Y is not None:
        return X, Y, num_batches
    else:
        return X, num_batches


def range_str(range_str, sort=True):
    """Generate range of ints from a formatted string,
    then convert range from int to str

    Example:
        >>> range_str('1-4,6,9-11')
        ['1','2','3','4','6','9','10','11']

    Takes a range in form of "a-b" and returns
    a list of numbers between a and b inclusive.
    Also accepts comma separated ranges like "a-b,c-d,f"  which will
    return a list with numbers from a to b, c to d, and f.

    Parameters
    ----------
    range_str : str
        of form 'a-b,c'
        where a hyphen indicates a range
        and a comma separates ranges or single numbers
    sort : bool
        If True, sort output before returning. Default is True.

    Returns
    -------
    list_range : list
        of int, produced by parsing range_str
    """
    # adapted from
    # http://code.activestate.com/recipes/577279-generate-list-of-numbers-from-hyphenated-and-comma/
    s = "".join(range_str.split())  # removes white space
    list_range = []
    for substr in range_str.split(','):
        subrange = substr.split('-')
        if len(subrange) not in [1, 2]:
            raise SyntaxError("unable to parse range {} in labelset {}."
                              .format(subrange, substr))
        list_range.extend(
            [int(subrange[0])]
        ) if len(subrange) == 1 else list_range.extend(
            range(int(subrange[0]), int(subrange[1]) + 1))

    if sort:
        list_range.sort()

    return [str(list_int) for list_int in list_range]



from glob import glob
import os
import re

import numpy as np
import torch


def find_fname(fname, ext):
    """given a file extension, finds a filename with that extension within
    another filename. Useful to find e.g. names of audio files in
    names of spectrogram files.

    Parameters
    ----------
    fname : str
        filename to search for another filename with a specific extension
    ext : str
        extension to search for in filename

    Returns
    -------
    sub_fname : str or None


    Examples
    --------
    >>> sub_fname(fname='llb3_0003_2018_04_23_14_18_54.wav.mat', ext='wav')
    'llb3_0003_2018_04_23_14_18_54.wav'
    """
    m = re.match(f'[\S]*{ext}', fname)
    if hasattr(m, 'group'):
        return m.group()
    elif m is None:
        return m


def _files_from_dir(dir_path, ext):
    """helper function that gets all files with a given extension
    from a directory or its sub-directories.

    If no files with the specified extension are found in the directory, then
    the function recurses into all sub-directories and returns any files with
    the extension in those sub-directories.

    Parameters
    ----------
    dir_path : str
        path to target directory
    ext : str
        file extension to search for

    Returns
    -------
    files : list
        of paths to files with specified file extension

    Notes
    -----
    used by vak.io.audio.files_from_dir and vak.io.annot.files_from_dir
    """
    wildcard_with_extension = f'*.{ext}'
    files = sorted(
        glob(os.path.join(dir_path, wildcard_with_extension))
    )
    if len(files) == 0:
        # if we don't any files with extension, look in sub-directories
        files = []
        subdirs = glob(os.path.join(dir_path, '*/'))
        for subdir in subdirs:
            files.extend(
                glob(os.path.join(dir_path, subdir, wildcard_with_extension))
            )

    if len(files) == 0:
        raise FileNotFoundError(
            f'No files with extension {ext} found in '
            f'{dir_path} or immediate sub-directories'
        )

    return files


def timebin_dur_from_vec(time_bins, n_decimals_trunc=3):
    """compute duration of a time bin, given the
    vector of time bin centers associated with a spectrogram

    Parameters
    ----------
    time_bins : numpy.ndarray
        vector of times in spectrogram, where each value is a bin center.
    n_decimals_trunc : int
        number of decimal places to keep when truncating the timebin duration calculated from
        the spectrogram arrays.
        Default is 3, i.e. assumes milliseconds is the last significant digit.

    Returns
    -------
    timebin_dur : float
        duration of a timebin, estimated from vector of times

    Notes
    -----
    takes mean of pairwise difference between neighboring time bins,
    to deal with floating point error, then rounds and truncates to specified decimal place
    """
    # first we round to the given number of decimals
    timebin_dur = np.around(
        np.mean(np.diff(time_bins)),
        decimals=n_decimals_trunc
    )
    # only after rounding do we truncate any decimal place past decade
    decade = 10 ** n_decimals_trunc
    timebin_dur = np.trunc(timebin_dur * decade) / decade
    return timebin_dur


def safe_truncate(X, Y, spect_ID_vector, labelmap, target_dur, timebin_dur):
    correct_length = np.round(target_dur / timebin_dur).astype(int)
    if X.shape[-1] == correct_length:
        return X, Y, spect_ID_vector
    elif X.shape[-1] > correct_length:
        # truncate from the front
        X_out = X[:, :correct_length]
        Y_out = Y[:correct_length, :]
        spect_ID_vector_out = spect_ID_vector[:correct_length]

        mapvals = np.asarray(
            sorted(list(labelmap.values()))
        )

        if np.array_equal(np.unique(Y_out), mapvals):
            pass
        else:
            # if all classes weren't in truncated arrays,
            # try truncating from the back instead
            X_out = X[:, -correct_length:]
            Y_out = Y[-correct_length:, :]
            spect_ID_vector_out = spect_ID_vector[:correct_length]

            if not np.array_equal(np.unique(Y_out), mapvals):
                raise ValueError(
                    "was not able to truncate in a way that maintained all classes in dataset"
                )

        return X_out, Y_out, spect_ID_vector_out

    elif X.shape[-1] < correct_length:
        raise ValueError(
            f"arrays have length {X.shape[-1]} that is shorter than correct length, {correct_length}, "
            f"(= target duration {target_dur} / duration of timebins, {timebin_dur})."
        )


def get_default_device():
    """get default device for torch.

    Returns
    -------
    device : str
        'cuda' if torch.cuda.is_available() is True,
        and returns 'cpu' otherwise.
    """
    if torch.cuda.is_available():
        return "cuda"
    else:
        return "cpu"



import os
from pathlib import Path

import crowsetta

from ..config import validators
from ..util.general import _files_from_dir


def files_from_dir(annot_dir, annot_format):
    """get all annotation files of a given format
    from a directory or its sub-directories,
    using the file extension associated with that annotation format.
    """
    if annot_format not in validators.VALID_ANNOT_FORMATS:
        raise ValueError(
            f'specified annotation format, {annot_format} not valid.\n'
            f'Valid formats are: {VALID_ANNOT_FORMATS}'
        )

    format_module = getattr(crowsetta.formats, annot_format)
    ext = format_module.meta.ext
    annot_files = _files_from_dir(annot_dir, ext)
    return annot_files


def _recursive_stem(path_str):
    name = Path(path_str).name
    stem, ext = os.path.splitext(name)
    ext = ext.replace('.', '')
    while ext not in validators.VALID_AUDIO_FORMATS:
        new_stem, ext = os.path.splitext(stem)
        ext = ext.replace('.', '')
        if new_stem == stem:
            raise ValueError(
                f'unable to compute stem of {path_str}'
            )
        else:
            stem = new_stem
    return stem


def source_annot_map(source_files, annot_list):
    """map annotations to their source files, i.e. audio or spectrogram files

    Parameters
    ----------
    source_files : list
        of audio or spectrogram files. The names of the files must match the
        file attribute of the annotations. E.g., if an audio file is
        'bird0-2016-05-04-133027.wav', then there must be an annotation whose
        file attribute equals that filename. Spectrogram files should include
        the audio file name, e.g. 'bird0-2016-05-04-133027.wav.mat' or
        'bird0-2016-05-04-133027.spect.npz'
    annot_list : list
        of Annotations corresponding to files in source_files
    """
    # to pair audio files with annotations, make list of tuples
    source_annot_map = []  # that we convert a dict before returning
    # we copy source files so we can convert to stem, and so we can
    # pop items off copy without losing original list.
    # We pop to validate that function worked, by making sure there are
    # no items left in this list after the loop
    source_files_stem = source_files.copy()
    source_files_stem = [_recursive_stem(sf) for sf in source_files_stem]
    source_file_inds = list(range(len(source_files)))
    for annot in annot_list:
        # remove stem so we can find .spect files that match with audio files,
        # e.g. find 'llb3_0003_2018_04_23_14_18_54.mat' that should match
        # with 'llb3_0003_2018_04_23_14_18_54.wav'
        annot_file_stem = _recursive_stem(annot.audio_file)

        ind_in_stem = [ind
                       for ind, source_file_stem in enumerate(source_files_stem)
                       if annot_file_stem == source_file_stem]
        if len(ind_in_stem) > 1:
            more_than_one = [source_file_inds[ind] for ind in ind_in_stem]
            more_than_one = [source_files[ind] for ind in more_than_one]
            raise ValueError(
                "Found more than one source file that matches an annotation."
                f"\nSource files are: {more_than_one}."
                f"\nAnnotation has file set to '{annot.audio_file}' and is: {annot}"
            )
        elif len(ind_in_stem) == 0:
            raise ValueError(
                "Did not find a source file matching the following annotation: "
                f"\n{annot}. Annotation has file set to '{annot.audio_file}'."
            )
        else:
            ind_in_stem = ind_in_stem[0]
            ind = source_file_inds[ind_in_stem]
            source_annot_map.append(
                (source_files[ind], annot)
            )
            source_files_stem.pop(ind_in_stem)
            source_file_inds.pop(ind_in_stem)

    if len(source_files_stem) > 0:
        raise ValueError(
            'could not map the following source files to annotations: '
            f'{source_files_stem}'
        )

    return dict(source_annot_map)



import logging
import os
import shutil
import sys
from datetime import datetime
from configparser import ConfigParser

import numpy as np


def learning_curve(train_vds_path,
                   test_vds_path,
                   total_train_set_duration,
                   train_set_durs,
                   num_replicates,
                   networks,
                   num_epochs,
                   config_file,
                   val_vds_path=None,
                   val_error_step=None,
                   checkpoint_step=None,
                   patience=None,
                   save_only_single_checkpoint_file=True,
                   normalize_spectrograms=False,
                   use_train_subsets_from_previous_run=False,
                   previous_run_path=None,
                   root_results_dir=None,
                   save_transformed_data=False,
                   ):
    """generate learning curve, by training models on training sets across a
    range of sizes and then measure accuracy of those models on a test set

    Parameters
    ----------
    train_vds_path : str
        path to Dataset that represents training data
    test_vds_path : str
        path to Dataset that represents test data
    total_train_set_duration : int
        total duration of training set, in seconds
    train_set_durs : list
        of int, durations in seconds of subsets taken from training data
        to create a learning curve, e.g. [5, 10, 15, 20]
    num_replicates : int
        number of times to replicate training for each training set duration
        to better estimate mean accuracy for a training set of that size.
        Each replicate uses a different randomly drawn subset of the training
        data (but of the same duration).
    networks : dict
        where each key is the name of a neural network and the corresponding
        value is the configuration for that network (in a namedtuple or a dict)
    num_epochs : int
        number of training epochs. One epoch = one iteration through the entire
        training set.
    config_file : str
        path to config.ini file. Used to rewrite file with options determined by
        this function and needed for other functions (e.g. cli.summary)
    val_vds_path : str
        path to Dataset that represents validation data.
        Default is None, in which case accuracy is not measured on validation set
        during training.
    val_error_step : int
        step/epoch at which to estimate accuracy using validation set.
        Default is None, in which case accuracy is not measured on validation set
        during training.
    checkpoint_step : int
        step/epoch at which to save to checkpoint file.
        Default is None, in which case checkpoint is only saved at the last epoch.
    patience : int
        number of epochs to wait without the error dropping before stopping the
        training. Default is None, in which case training continues for num_epochs
    save_only_single_checkpoint_file : bool
        if True, save only one checkpoint file instead of separate files every time
        we save. Default is True.
    normalize_spectrograms : bool
        if True, use utils.spect.SpectScaler to normalize the spectrograms.
        Normalization is done by subtracting off the mean for each frequency bin
        of the training set and then dividing by the std for that frequency bin.
        This same normalization is then applied to validation + test data.
    use_train_subsets_from_previous_run : bool
        if True, use training subsets saved in a previous run
    previous_run_path : str
        path to results directory from a previous run
    root_results_dir : str
        path in which to create results directory for this run of cli.learncurve
    save_transformed_data : bool
        if True, save transformed data (i.e. scaled, reshaped). The data can then
        be used on a subsequent run of learncurve (e.g. if you want to compare results
        from different hyperparameters across the exact same training set).
        Also useful if you need to check what the data looks like when fed to networks.

    Returns
    -------
    None

    Saves results in root_results_dir and adds some options to config_file.
    """
    # ---------------- pre-conditions ----------------------------------------------------------------------------------
    if val_error_step and val_vds_path is None:
        raise ValueError(
            f"val_error_step set to {val_error_step} but val_vds_path is None; please provide a path to "
            f"a validation data set that can be used to check error rate very {val_error_step} steps"
        )

    if val_vds_path and val_error_step is None:
        raise ValueError(
            "val_vds_path was provided but val_error_step is None; please provide a value for val_error_step"
        )

    max_train_set_dur = np.max(train_set_durs)
    if max_train_set_dur > total_train_set_duration:
        raise ValueError('Largest duration for a training set of {} '
                         'is greater than total duration of training set, {}'
                         .format(max_train_set_dur, total_train_set_duration))

    # need to set up a results dir so we have some place to put the log file
    timenow = datetime.now().strftime('%y%m%d_%H%M%S')
    results_dirname = f'{timenow}'
    if root_results_dir:
        results_dirname = os.path.join(root_results_dir,
                                       results_dirname)
    os.makedirs(results_dirname)
    shutil.copy(config_file, results_dirname)

    # ---------------- logging -----------------------------------------------------------------------------------------
    logfile_name = os.path.join(results_dirname,
                                'logfile_from_running_learncurve_' + timenow + '.log')
    # give logger same name that core.learncurve will use, so it will have the same level and handlers
    # as what's declared here
    logger = logging.getLogger('learncurve')
    logger.setLevel('INFO')
    logger.addHandler(logging.FileHandler(logfile_name))
    logger.addHandler(logging.StreamHandler(sys.stdout))
    logger.info('Logging results to {}'.format(results_dirname))
    logger.info('Using config file: {}'.format(config_file))

    # --------------- let core.learncurve do all the work --------------------------------------------------------------

    # rewrite config file,
    # so that paths where results were saved are automatically in config
    config = ConfigParser()
    config.read(config_file)
    config.set(section='LEARNCURVE',
               option='results_dir_made_by_main_script',
               value=results_dirname)
    with open(config_file, 'w') as config_file_rewrite:
        config.write(config_file_rewrite)
